from atom.api import Str, Typed, ForwardTyped, set_default
from enaml.core.declarative import d_
from enaml.widgets.control import Control, ProxyControl


class ProxyFontComboBox(ProxyControl):
    """ The abstract definition of a proxy FontComboBox object.

    """
    #: A reference to the FontComboBox declaration.
    declaration = ForwardTyped(lambda: FontComboBox)

    def set_font(self, font: str):
        raise NotImplementedError


class FontComboBox(Control):
    """ A drop-down list from which a font can be selected at a time.

    Fonts are pulled from the system.
    """
    #: The font family currently selected.
    font = d_(Str())

    #: A read only cached property which returns the list of fonts
    # fonts = d_(Property(cached=True), writable=False)

    #: A FontComboBox hugs its width weakly by default.
    hug_width = set_default('weak')

    #: A reference to the ProxyFontComboBox object.
    proxy = Typed(ProxyFontComboBox)



from collections import namedtuple
from dataclasses import dataclass
from decimal import Decimal

import pandas as pd
import pytest

from enamlext.qt.qt_dataframe import DataFrameProxy
from enamlext.qt.table.column import Column, generate_columns, Alignment


def test_get_displayed_value():
    class Foo:
        def __str__(self) -> str:
            return 'string for Foo'

        def __format__(self, spec: str) -> str:
            s = str(self)
            if spec == 'u':
                s = s.upper()
            else:
                s = format(s, spec)
            return s

    foo = Foo()

    item = {'foo': foo}

    column = Column('foo', use_getitem=True)

    assert foo is column.get_value(item)
    assert 'string for Foo' == column.get_displayed_value(item)

    column2 = Column('foo', use_getitem=True, fmt='u')
    assert 'STRING FOR FOO' == column2.get_displayed_value(item)


def test_giving_fmt_as_callback():
    def uppercase_mirrored(value):
        return str(value)[::-1].upper()

    class InterestingClass:
        def __str__(self): return 'TextForUsers'

    obj = InterestingClass()
    item = {'obj': obj, 'foo': None}

    column = Column('obj', fmt=uppercase_mirrored, use_getitem=True)

    assert 'SRESUROFTXET' == column.get_displayed_value(item)



def test_generate_columns_with_hints():
    @dataclass
    class Trade:
        symbol: str
        price: Decimal
        quantity: int

    items = [
        t1 := Trade(symbol='FOO', price=15.2, quantity=500),
        t2 := Trade(symbol='BAR', price=22.3, quantity=800),
        t3 := Trade(symbol='FOO', price=16.1, quantity=100),
    ]

    hints = {
        'symbol': {
            'title': 'Ticker',
        },
        'price': {
            'fmt': ',.2f',
        }
    }
    generated_columns = generate_columns(items, hints=hints, exclude=['quantity'])

    column_symbol, column_price = generated_columns

    assert column_symbol.title == 'Ticker'

    assert column_price.title == 'Price'
    assert column_price.fmt == ',.2f'

    assert '22.30' == column_price.get_displayed_value(t2)


def test_generate_columns_with_pandas_dataframe():
    df = pd.DataFrame()
    df['symbol'] = ['A', 'B']
    df['price'] = [10.25, 12.0]

    generated_columns = generate_columns(DataFrameProxy(df))

    column_symbol, column_price = generated_columns

    assert column_symbol.title == 'Symbol'
    assert column_price.title == 'Price'
    assert column_price.align == Alignment.RIGHT


df = pd.DataFrame()
df['symbol'] = ['A', 'B']
df['price'] = [10.25, 12.0]
df['currency'] = ['EUR', 'GBP']

Record = namedtuple('Record', 'symbol price currency')
named_tuples = [Record('A', 10.25, 'EUR'), Record('B', 12.0, 'GBP')]

tuples = [('A', 10.25, 'EUR'), ('B', 12.0, 'GBP')]

@dataclass
class RecordData:
    symbol: str
    price: float
    currency: str

records = [RecordData('A', 10.25, 'EUR'), RecordData('B', 12.0, 'GBP')]

dicts = [{'symbol': 'A', 'price': 10.25, 'currency': 'EUR'},
         {'symbol': 'B', 'price': 12.0, 'currency': 'GBP'}, ]

@pytest.mark.parametrize(
    ['items', 'include', 'hints'],
    [
        (DataFrameProxy(df), ['currency', 'price'], {}),
        (named_tuples, ['currency', 1], {}),
        (tuples, [2, 1], {2: {'title': 'Currency'}, 1: {'title': 'Price'}}),
        (records, ['currency', 'price'], {}),
        (dicts, ['currency', 'price'], {}),
    ]
)
def test_generate_columns_include(items, include, hints):
    col_1, col_2 = generate_columns(items, include=include, hints=hints)

    def assert_column(column, title, align):
        assert title == column.title
        assert align == column.align

    assert_column(col_1, 'Currency', Alignment.LEFT)
    assert_column(col_2, 'Price', Alignment.RIGHT)


@pytest.mark.parametrize(
    ['items', 'exclude', 'hints'],
    [
        (DataFrameProxy(df), ['price'], {}),
        (named_tuples, ['price'], {}),
        (tuples, [1], {0: {'title': 'Symbol'}, 2: {'title': 'Currency'}}),
        (records, ['price'], {}),
        (dicts, ['price'], {}),
    ]
)
def test_generate_columns_exclude(items, exclude, hints):
    col_1, col_2 = generate_columns(items, exclude=exclude, hints=hints)

    def assert_column(column, title, align):
        assert title == column.title
        assert align == column.align

    assert_column(col_1, 'Symbol', Alignment.LEFT)
    assert_column(col_2, 'Currency', Alignment.LEFT)



from typing import List, Any, Optional

from atom.api import Int, Typed
from enaml.qt.qt_control import QtControl

from enamlext.qt.qtable import QTable, DoubleClickContext, SelectionContext, SelectionMode
from enamlext.qt.table.column import Column
from enamlext.qt.table.summary import TableSelectionSummary
from enamlext.widgets.table import ProxyTable

# cyclic notification guard flags
INDEX_GUARD = 0x1


SELECTION_MODES_MAP = {
    'cell': SelectionMode.SINGLE_CELL,
    'cells': SelectionMode.MULTI_CELLS,
    'row': SelectionMode.SINGLE_ROW,
    'rows': SelectionMode.MULTI_ROWS,
}


class QtTable(QtControl, ProxyTable):
    """ A Qt implementation of an Enaml Table.
    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QTable)

    #: Cyclic notification guard. This a bitfield of multiple guards.
    _guard = Int(0)

    # Initialization API
    def create_widget(self):
        """ Create the QTable widget.
        """
        if hasattr(self.declaration, 'convert_item'):
            convert_item = self.declaration.convert_item
        else:
            convert_item = None

        self.widget = QTable([], parent=self.parent_widget(),
                             convert_item=convert_item)

    def init_widget(self):
        """ Create and initialize the underlying widget.

        """
        super().init_widget()
        d = self.declaration
        with self.widget.updating_internals():
            self.set_columns(d.columns)
            self.set_items(d.items)
            self.set_selected_items(d.selected_items)
            self.set_context_menu(d.context_menu)
            self.set_checkable(d.checkable)
            self.set_sortable(d.sortable)
            self.set_selection_mode(d.selection_mode)

        # double click action
        self.widget.on_double_click.connect(self._on_double_clicked)

        # single click selection
        self.widget.on_selection.connect(self._on_selection_changed)
        # self.widget.currentIndexChanged.connect(self.on_index_changed)

    # Signal Handlers
    def _on_double_clicked(self, context: DoubleClickContext):
        # TODO: DoubleClickContext has a lot of knowledge of Qt details - we don't want this to leak!
        self.declaration.double_clicked(context)

    def _on_selection_changed(self, context: SelectionContext):
        # TODO: think better if this convertion for selected_items should be implemented
        #       inside the QTable (qtable.py)
        selected_items = context.selected_items
        if hasattr(self.declaration, 'convert_item'):
            selected_items = [self.declaration.convert_item(item) for item in selected_items]
        self.declaration.selected_items = selected_items
        self.declaration.selection_changed(context)

        if self.declaration.show_summary:
            self.refresh_summary(context)

    # ProxyTable API
    def set_items(self, items: List[Any]):
        """ Set the items (rows) of the QTable.
        """
        with self.widget.updating_internals():
            self.widget.items = items

    def set_columns(self, columns: List[Column]):
        """ Set the columns of the QTable.
        """
        with self.widget.updating_internals():
            self.widget.columns = columns

    def set_selected_items(self, selected_items: List[Any]):
        pass

    def set_context_menu(self, context_menu: List):
        self.widget.context_menu = context_menu # TODO: make this a property of the QTable

    def set_checkable(self, checkable: bool) -> None:
        self.widget.checkable = checkable

    def set_sortable(self, sortable: bool) -> None:
        self.widget.sortable = sortable

    def set_selection_mode(self, selection_mode: str) -> None:
        self.widget.set_selection_mode(SELECTION_MODES_MAP[selection_mode])

    def set_hints(self, hints: dict) -> None:
        ...  # do nothing because underlying widget does not know about hints

    def refresh_summary(self, context: Optional[SelectionContext] = None) -> None:
        if context is None:
            context = self.widget.get_current_selection_context()
        self.declaration.summary = TableSelectionSummary.from_selection_context(context)



from atom.api import (
    Typed, ForwardTyped, Int, observe
)

from enaml.core.declarative import d_

from enaml.widgets.control import Control, ProxyControl



class ProxyStarRating(ProxyControl):
    """ The abstract definition of a proxy StarRatingWidget object.

    """
    #: A reference to the StarRating declaration.
    declaration = ForwardTyped(lambda: StarRating)

    def set_stars(self, stars):
        raise NotImplementedError

    def set_rating(self, rating):
        raise NotImplementedError



class StarRating(Control):
    """ Enaml declarative control for giving a Star Rating widget.

    """
    stars = d_(Int(default=5))
    rating = d_(Int())

    #: A reference to the ProxyStarRating object.
    proxy = Typed(ProxyStarRating)

    #--------------------------------------------------------------------------
    # Observers
    #--------------------------------------------------------------------------
    @observe('stars', 'rating')
    def _update_proxy(self, change):
        """ An observer which sends the state change to the proxy.

        """
        # The superclass implementation is sufficient.
        super(StarRating, self)._update_proxy(change)


def star_rating_factory():
    from enamlext.qt.qt_star import QtStarRating
    return QtStarRating


from enaml.qt.qt_factories import QT_FACTORIES
QT_FACTORIES['StarRating'] = star_rating_factory


from atom.api import Atom, Str
from enaml.widgets.api import Field

from enamlext.auto import create_editor
from enamlext.auto._editor import connect, Editor


class Person(Atom):
    name = Str()


def test_simple_editor():
    person = Person()
    editor = create_editor(person)

    assert 2 == len(editor.children)


def test_exclude():
    class Person2(Atom):
        name = Str().tag(exclude=True)

    person = Person2()
    editor = create_editor(person)

    assert 0 == len(editor.children)


def test_connect():
    person = Person()
    field = Field()
    connect(person, 'name', field, 'text')

    assert field.text == ''
    assert person.name == ''

    # Update the model causes the view to update
    person.name = 'John'
    assert field.text == 'John'

    # Update the view causes the model to update
    field.text = 'Paul'
    assert person.name == 'Paul'


def test_initial_sync():
    person = Person(name='George')
    editor = create_editor(person)

    _, field = editor.children

    assert field.text == 'George'


def test_label():
    class Person2(Atom):
        name = Str().tag(label='Full Name')

    p = Person2()
    editor = create_editor(p)

    label, _ = editor.children
    assert label.text == 'Full Name'


def test_create_editor_control():
    person = Person(name='George')
    editor = Editor(obj=person)
    assert editor.obj is person
    assert len(editor.children) == 2




from collections import namedtuple
from dataclasses import dataclass

from enamlext.qt.table.summary import TableSelectionSummary, compute_summary
from enamlext.qt.table.column import Column, Alignment, generate_columns
from enamlext.qt.table.filtering import TableFilters, Filter


def test_generate_columns_from_list_of_tuples():
    items = [('John', 10), ('Paul', 20)]
    col_1, col_2 = generate_columns(items)

    assert ['0', '1'] == [col_1.title, col_2.title]

    assert col_1.align == Alignment.LEFT
    assert col_2.align == Alignment.RIGHT

    assert ['John', 'Paul'] == [col_1.get_value(item) for item in items]
    assert [10, 20] == [col_2.get_value(item) for item in items]


def test_generate_columns_from_list_of_dicts():
    items = [{'name': 'George', 'age': 30}, {'name': 'Ringo', 'age': 45}]
    col_1, col_2 = generate_columns(items)

    assert ['Name', 'Age'] == [col_1.title, col_2.title]

    assert col_1.align == Alignment.LEFT
    assert col_2.align == Alignment.RIGHT

    assert ['George', 'Ringo'] == [col_1.get_value(item) for item in items]
    assert [30, 45] == [col_2.get_value(item) for item in items]


def test_generate_columns_from_list_of_dataclass_instances():
    @dataclass
    class Person:
        name: str
        age: int

    items = [Person(name='Alice', age=6), Person(name='Sophie', age=12)]
    col_1, col_2 = generate_columns(items)

    assert ['Name', 'Age'] == [col_1.title, col_2.title]

    assert col_1.align == Alignment.LEFT
    assert col_2.align == Alignment.RIGHT

    assert ['Alice', 'Sophie'] == [col_1.get_value(item) for item in items]
    assert [6, 12] == [col_2.get_value(item) for item in items]


def test_generate_columns_from_list_of_namedtuples():
    Person = namedtuple('Person', ['age', 'name'])

    items = [Person(age=66, name='Bob'), Person(age=99, name='Charlie')]
    col_1, col_2 = generate_columns(items)

    assert ['Age', 'Name'] == [col_1.title, col_2.title]

    assert col_1.align == Alignment.RIGHT
    assert col_2.align == Alignment.LEFT

    assert [66, 99] == [col_1.get_value(item) for item in items]
    assert ['Bob', 'Charlie'] == [col_2.get_value(item) for item in items]


############################
# Filter
############################


def test_filter():
    column = Column('age', use_getitem=True)

    filter = Filter(column, '> 25')

    assert filter({'age': 30})
    assert not filter({'age': 25})
    assert not filter({'age': 20})


def test_filter_assumes_equality_by_default():
    column = Column('age', use_getitem=True)

    filter = Filter(column, '25')

    assert filter({'age': 25})
    assert not filter({'age': 30})
    assert not filter({'age': 20})


def test_filter_str():
    filter = Filter(None, '> 15')
    assert str(filter) == '> 15'


def test_filters():
    col_age = Column('age', use_getitem=True)
    col_name = Column('name', use_getitem=True)

    filters = TableFilters([
        Filter(col_age, '> 12'),
        Filter(col_name, 'Leo'),
    ])

    items = [
        {'name': 'Leo', 'age': 13},
        {'name': 'Leo', 'age': 11},
        {'name': 'Leo', 'age': 18},
        {'name': 'Joe', 'age': 11},
        {'name': 'Joe', 'age': 14},
        {'name': 'Joe', 'age': 20},
        {'name': 'Leo', 'age': 19},
    ]

    expected = [
        {'name': 'Leo', 'age': 13},
        {'name': 'Leo', 'age': 18},
        {'name': 'Leo', 'age': 19},
    ]

    assert expected == list(filters.filter_items(items))


def test_get_filter_for_column():
    col_1 = Column('a')
    col_2 = Column('b')

    filters = TableFilters([Filter(col_1, '> 10'), Filter(col_2, '<= 12')])

    assert str(filters.get(col_1)) == '> 10'
    assert str(filters.get(col_2)) == '<= 12'


###########
# Summary
###########

def test_compute_summary():
    values = [1, 2, 'foo', 'bar', -3, -4, -1]
    summary = compute_summary(values)
    expected_summary = TableSelectionSummary(sum=-5, values=values, min=-4, max=2,
                                             count_numbers=5)
    assert expected_summary == summary

    assert 7 == summary.count
    assert -1 == summary.avg


def test_compute_summary_diff():
    values = [-4, 11]
    summary = compute_summary(values)
    expected_summary = TableSelectionSummary(sum=7, values=values, min=-4, max=11,
                                             count_numbers=2)
    assert expected_summary == summary

    assert 2 == summary.count
    assert 3.5 == summary.avg
    assert 15 == summary.diff


def test_summary_string_text():
    summary = TableSelectionSummary(sum=5, values=[...] * 7, min=-4.5, max=12.25, count_numbers=4)
    assert 'Count: 7   Average: 1.25   Sum: 5   CountNumbers: 4   Min: -4.5   Max: 12.25' == str(summary)


def test_summary_string_text_with_diff():
    summary = TableSelectionSummary(sum=4, values=[1, 3], min=1, max=3, count_numbers=2)
    assert 'Count: 2   Average: 2.0   Sum: 4   CountNumbers: 2   Min: 1   Max: 3   Diff: 2' == str(summary)



from enum import Enum
from typing import TypedDict, Optional, Callable

from PyQt5.QtGui import QColor


class CellStyle(TypedDict, total=False):
    color: Optional[QColor]
    background: Optional[QColor]
    font: Optional[str]


CellStyleCallback = Callable[['TableContext'], CellStyle]


class ColumnSize(str, Enum):
    AUTO = 'auto'
    JUST = 'just'  # based on the font metrics of the table view



from operator import attrgetter
from typing import Dict

from atom.api import Atom
from atom.atom import AtomMeta, observe
from atom.catom import Member
from atom.typing_utils import ChangeDict
from enaml.application import deferred_call


def _iter_atom_members(dct: Dict):
    for k, v in dct.items():
        if issubclass(type(v), Member):
            yield k, v


def _create_refresh_method(name: str, bind: str):
    first_part, _ = bind.split('.')
    def _refresh(self, change: ChangeDict) -> None:
        if attrgetter(first_part)(self) is not None:
            deferred_call(setattr, self, name, attrgetter(bind)(self))
    _refresh.__name__ = f'_refresh_{name}'
    return observe(bind, first_part)(_refresh)


class ViewModelMeta(AtomMeta):
    def __new__(metacls, name, bases, dct):
        for name, value in _iter_atom_members(dct.copy()):
            if value.metadata and (bind := value.metadata.get('bind')):
                refresh_func = _create_refresh_method(name, bind)
                dct[f'_refresh_{name}'] = refresh_func

        return super().__new__(metacls, name, bases, dct)


class ViewModel(Atom, metaclass=ViewModelMeta):
    """
    Base class for models that are directly bound to an Enaml view.

    An example:

    Suppose you have a Controller with some business logic like:

    class Controller(Atom):
        orders = List()

    and you have a view model like:

    class Model(Atom):
        controller = Typed(Controller)
        orders = List()

        @observe('controller.orders')
        def _refresh_orders(self, change):
            deferred_call(setattr, self, 'orders', self.controller.orders)

    and then in your enaml UI you have:

    enamldef OrdersWidget(Container):
        Table:
            items << model.orders

    So, basically, for every member like 'orders' above you would have to write
    its own wrapper method ensuring that the refresh of the UI is taken care via
    deferred calls. That's a lot of boilerplate.

    Now compare with:

    class Controller(Atom):
        orders = List()

    class Model(ViewModel):
        controller = Typed(Controller)
        orders = List().tag(bind='controller.orders')


    Job done! The @observe methods will be automatically generated on class initialization,
    and they will ensure that anything that gets set and is bound to a control in the UI
    will be set using deferred_call.
    """



from atom.api import Atom, List, Str, Typed
from enamlext.view_model import ViewModel


def test_view_model_subclassing(mocker):
    mock_deferred_call = mocker.patch('enamlext.view_model.deferred_call', autospec=True)

    class Controller(Atom):
        orders = List()
        other = Str()

    class OrderViewModel(ViewModel):
        controller = Typed(Controller)
        orders = List().tag(bind='controller.orders')  # bound to the view - set with deferred_call
        yet_another = Str()

    controller = Controller()
    view_model = OrderViewModel(controller=controller)

    controller.orders = ['Order1', 'Order2']
    controller.other = 'other'
    view_model.yet_another = 'yet_another'

    mock_deferred_call.assert_has_calls([
        # set controller
        mocker.call(setattr, view_model, 'orders', []),
        # set controller.orders (blank)
        mocker.call(setattr, view_model, 'orders', []),
        # finally the set controller.orders with orders
        mocker.call(setattr, view_model, 'orders', ['Order1', 'Order2']),
   ])


"""
MultiSelect widget - allows for multiple selection/choices.
"""
from typing import Optional
from unittest import mock

import pytest
from atom.api import Atom, List
from atom.atom import observe
from atom.property import Property


def is_valid_selection(items: List, selected_items: List) -> bool:
    selected_items_set = set(selected_items)
    items_set = set(items)
    return ((selected_items_set <= items_set)  # subset
            and (len(selected_items_set) == len(selected_items))  # no dupes
            and (len(items_set) == len(items)))


class MultiSelectInitError(ValueError):
    pass


class MultiSelectUpdateError(RuntimeError):
    pass


class MultiSelectModel(Atom):
    #: collection of all available items for selection
    items = List()

    #: collection of currently selected items
    selected_items = List()

    #: collection of items still up for grabs (read-only!)
    available_items = Property(cached=True)

    def __init__(self, items: List, selected_items: Optional[List] = None):
        """
        The list of all available items must always be provided.
        Optionally the model can also be initialized with some selected_items.
        The list of available items then it's always calculated in terms of the other two above.
        """
        if selected_items is None:
            selected_items = []
        if not is_valid_selection(items, selected_items):
            raise MultiSelectInitError(f"{type(self).__name__} initialized with invalid lists. "
                                       f"The selected_items must bea a subset of all the items and "
                                       f"there must be no duplicates.")
        with self.suppress_notifications():
            self.items = items

        # Now, finally, we fire the notifications
        self.selected_items = selected_items

    @observe("items", "selected_items")
    def _refresh_available_items(self, change=None):
        if not is_valid_selection(self.items, self.selected_items):
            raise MultiSelectUpdateError(f"{type(self).__name__} invalid state after an update. "
                                         f"The selected_items must bea a subset of all the items and "
                                         f"there must be no duplicates.")
        self.get_member("available_items").reset(self)

    @available_items.getter
    def get_available_items(self):
        return [i for i in self.items if i not in self.selected_items]

    def invert_selection(self):
        self.selected_items = self.available_items

    def select_all(self):
        self.selected_items = self.items

    def select_none(self):
        self.selected_items = []

    def select(self, items):
        # TODO: preserve the order of the original list of items
        selected_items = list(self.selected_items)
        selected_items.extend(items)
        self.selected_items = selected_items

    def unselect(self, items):
        items = set(items)
        self.selected_items = [item for item in self.selected_items if item not in items]


    clear_selection = select_none


def test_is_valid_selection():
    # there should be no duplicated items
    # the items should respect original order/sequence
    assert is_valid_selection([1, 2, 3], [1, 3])
    assert is_valid_selection([1, 2, 3], [])
    assert is_valid_selection([1, 2, 3], [1, 2, 3])
    assert not is_valid_selection([1, 2, 3], [1, 2, 3, 4])
    assert not is_valid_selection([1, 2, 3], [1, 2, 5])


def test_multi_select_model_refreshes_available_items_correctly():
    multi_select_model = MultiSelectModel(items=[1, 2, 3, 4], selected_items=[2, 3])
    assert multi_select_model.available_items == [1, 4]

    multi_select_model.items = [1, 2, 3, 4, 5]
    assert multi_select_model.available_items == [1, 4, 5]

    multi_select_model.selected_items = [1, 2, 3, 4, 5]
    assert multi_select_model.available_items == []


def test_can_observe_on_available_items_notifications():
    multi_select_model = MultiSelectModel(items=[1, 2, 3, 4], selected_items=[2, 3])

    callback = mock.Mock(spec=True)

    multi_select_model.observe("available_items", callback)

    multi_select_model.selected_items = [2, 3, 4]

    callback.assert_called_once()


def test_exception_is_raised_when_model_reaches_invalid_state_after_update():
    multi_select_model = MultiSelectModel(items=[1, 2, 3, 4], selected_items=[2, 3])
    with pytest.raises(MultiSelectUpdateError):
        multi_select_model.selected_items = [2, 2, 3]


def test_invert_selection():
    multi_select_model = MultiSelectModel(items=[1, 2, 3, 4], selected_items=[2, 3])
    assert multi_select_model.available_items == [1, 4]

    multi_select_model.invert_selection()
    assert multi_select_model.available_items == [2, 3]
    assert multi_select_model.selected_items == [1, 4]




from enaml.widgets.api import Field as _Field
from atom.api import Enum, List
from enaml.core.declarative import d_


class Field(_Field):
    submit_triggers = d_(List(
        Enum('lost_focus', 'return_pressed', 'auto_sync'),
        ['lost_focus', 'return_pressed', 'auto_sync']
    ))



import contextlib
import csv
import itertools
import logging
import math
import operator
import warnings
import weakref
from abc import abstractmethod, ABC
from dataclasses import dataclass
from enum import Enum, auto
from functools import lru_cache
from io import StringIO
from typing import Optional, Any, List, Tuple, NamedTuple, Collection, Set, Iterable

# Constants

from enamlext.qt.table.column import Column, Alignment, AUTO_ALIGN
from enamlext.qt.table.defs import CellStyle
from enamlext.qt.table.filtering import TableFilters, Filter
from qtpy.QtCore import QAbstractTableModel, QModelIndex, Qt, QObject, QPoint, Signal, QItemSelection, QEvent
from qtpy.QtGui import QContextMenuEvent, QFont, QColor, QPixmap, QKeySequence
from qtpy.QtWidgets import QApplication, QTableView, QMenu, QAction

from enamlext.qt.table.table_context import TableContext


logger = logging.getLogger(__name__)

DEFAULT_ROW_HEIGHT = 23
DEFAULT_FONT_NAME = "Calibri"
DEFAULT_FONT_SIZE_PX = 13

CHECKBOX_FLAG = Qt.ItemNeverHasChildren | Qt.ItemIsEditable | Qt.ItemIsUserCheckable | Qt.ItemIsEnabled


RED = QColor(Qt.red)


class SelectionMode(Enum):
    SINGLE_CELL = auto()
    MULTI_CELLS = auto()
    SINGLE_ROW = auto()
    MULTI_ROWS = auto()


QT_ALIGNMENT_MAP = {
    Alignment.LEFT: int(Qt.AlignLeft | Qt.AlignVCenter),
    Alignment.CENTER: int(Qt.AlignCenter | Qt.AlignVCenter),
    Alignment.RIGHT: int(Qt.AlignRight | Qt.AlignVCenter),
}

QtAlignment = int


def to_qt_alignment(align: Alignment) -> QtAlignment:
    return QT_ALIGNMENT_MAP[align]


@lru_cache(maxsize=256)
def make_custom_font(font_spec: str) -> QFont:
    font = QFont()
    tokens = font_spec.split()

    if 'bold' in tokens:
        font.setBold(True)
        tokens.remove('bold')

    if 'italic' in tokens:
        font.setItalic(True)
        tokens.remove('italic')

    if 'underline' in tokens:
        font.setItalic(True)
        tokens.remove('underline')

    for token in tokens[:]:
        if token.endswith('px'):
            font.setPixelSize(int(token[:-2]))
            tokens.remove(token)
            break
        elif token.endswith('pt'):
            font.setPointSize(int(token[:-2]))
            tokens.remove(token)
            break

    # everything else must be the font family
    if tokens:
        family = ' '.join(tokens)
        font.setFamily(family)

    return font


class Cell(NamedTuple):
    row: int
    column: int


def default_convert_item(item):
    return item


class QTableModel(QAbstractTableModel):

    #: signal used to notify the view whenever checked_items changes
    on_checked_items: Signal = Signal(set)

    def __init__(self,
                 columns: List[Column],
                 items: Optional[List[Any]] = None,
                 *,
                 checkable: bool = False,
                 checked_items: Optional[Collection[Any]] = None,
                 parent: Optional[QObject] = None,
                 error_handling: str = 'graceful',  # TODO: other modes
                 convert_item = default_convert_item,
                 ):
        super().__init__(parent)
        self.columns = columns
        self._original_items = items
        self.checkable = checkable
        self.error_handling = error_handling
        if convert_item is None:
            convert_item = default_convert_item
        self.convert_item = convert_item

        # Filtering
        self._filtered_items = None
        self.filters = TableFilters()
        self._apply_filters()

        # Internally we keep track of which items are checked using a set
        if checked_items is None:
            checked_items = set()
        if not isinstance(checked_items, set):
            warnings.warn(f"QTableModel checked_items is recommend to be a "
                          f"set, but got: {type(checked_items).__name__} instead - coercing to set")
            checked_items = set(checked_items)
        self.checked_items = checked_items

        self._font = self._create_font()  # cache the font
        self._font_bold = self._create_font(bold=True)

        self._last_sorting_column = None

    def __len__(self):
        return len(self.items)  # O(1)

    # QAbstractTableModel interface -----------------------------------------------------------------------------------

    def rowCount(self, parent: Optional[QModelIndex] = None) -> int:
        return len(self.items)  # O(1)

    def columnCount(self, parent: Optional[QModelIndex] = None) -> int:
        return len(self.columns) + int(self.checkable)  # O(1)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        column_index = index.column()
        if column_index == 0 and self.checkable:
            return CHECKBOX_FLAG
        else:
            return super().flags(index)

    def data(self, index: QModelIndex, role: int) -> Any:
        column = self.get_column_by_index(index.column())
        if column.collect_stats:
            t0 = time.perf_counter()
            result = self._data(index, role)
            t1 = time.perf_counter()
            column.record_stats(role=role, elapsed=t1 - t0)
            return result
        else:
            return self._data(index, role)

    def _data(self, index, role):
        if role == Qt.DisplayRole:
            if self.checkable:
                offset = 1
            else:
                offset = 0
            if (col_index := index.column()) or not self.checkable:
                column = self.columns[col_index - offset]  # O(1)
                item = self.items[index.row()]  # O(1)
                return column.get_displayed_value(item)
        elif role == Qt.TextAlignmentRole:
            if self.checkable:
                offset = 1
            else:
                offset = 0
            # Only the first column is checkable (index = 0) - so we need to account for that offset
            if (col_index := index.column()) or not self.checkable:
                column = self.columns[col_index - offset]  # O(1)
                item = self.convert_item(self.items[index.row()])  # O(1)
                try:
                    return to_qt_alignment(column.get_align(item))
                except Exception as exc:
                    if self.error_handling == 'graceful':
                        logger.warning(f'Error when resolving alignment for column: {col_index = }, '
                                       f'{column.title = !r}, {index.row() = }, {exc = }')
                    else:
                        raise
        elif role == Qt.ToolTipRole:
            col_index = index.column()
            column = self.get_column_by_index(col_index)
            context = TableContext(
                model=self,
                index=index,
                role=role,
                column_index=col_index,
                column=column,
                convert=self.convert_item,
            )  # TODO: should we check if the column has a tooltip callback before creating this?
            return column.get_tooltip(context)
        elif role == Qt.CheckStateRole and self.checkable and index.column() == 0:
            # TODO: think about using the check state of the vertical header to control checked items
            #       what if we want to have more than 1 column that is checkable?
            #       should we keep a collection of checked items separately or should we consider wrapping each
            #       item in a wrapper object that holds the checked state? how about a proxy object?
            item = self.items[index.row()]  # O(1)
            if item in self.checked_items:
                return Qt.Checked
            else:
                return Qt.Unchecked
        elif role == Qt.FontRole:
            col_index = index.column()
            column = self.get_column_by_index(col_index)
            if column.cell_style is not None:
                context = TableContext(
                    model=self,
                    index=index,
                    role=role,
                    column_index=col_index,
                    column=column,
                    convert=self.convert_item,
                )
                try:
                    style = column.get_cell_style(context)
                    if style is not None and (font_spec := style.get('font')) is not None:
                        return make_custom_font(font_spec)
                except Exception as exc:
                    if self.error_handling == 'graceful':
                        logger.warning(f'Error when resolving font for column: {col_index = }, '
                                       f'{column.title = !r}, {index.row() = }, {exc = }')
                    else:
                        raise

            return self._font

        elif role == Qt.ForegroundRole:
            col_index = index.column()
            column = self.get_column_by_index(col_index)
            if column.cell_style is not None:
                context = TableContext(
                    model=self,
                    index=index,
                    role=role,
                    column_index=col_index,
                    column=column,
                    convert=self.convert_item,
                )
                try:
                    style = column.get_cell_style(context)
                    if style is not None:
                        return style.get('color')
                except Exception as exc:
                    if self.error_handling == 'graceful':
                        logger.warning(f'Error when resolving foreground colour style for column: {col_index = }, '
                                       f'{column.title = !r}, {index.row() = }, {exc = }')
                    else:
                        raise

        elif role == Qt.BackgroundColorRole:
            col_index = index.column()
            column = self.get_column_by_index(col_index)
            if column.cell_style is not None:
                context = TableContext(
                    model=self,
                    index=index,
                    role=role,
                    column_index=col_index,
                    column=column,
                    convert=self.convert_item,
                )
                try:
                    style = column.get_cell_style(context)
                    if style is not None:
                        return style.get('background')
                except Exception as exc:
                    if self.error_handling == 'graceful':
                        logger.warning(f'Error when resolving background colour style for column: {col_index = }, '
                                       f'{column.title = !r}, {index.row() = }, {exc = }')
                    else:
                        raise

        elif role == Qt.DecorationRole:  # TODO: refactor (DRY)
            col_index = index.column()
            column = self.get_column_by_index(col_index)
            if column.image is not None:
                context = TableContext(
                    model=self,
                    index=index,
                    role=role,
                    column_index=col_index,
                    column=column,
                    convert=self.convert_item,
                )
                if (image := column.get_image(context)):
                    img = QPixmap()
                    img.load(image)

                    # TODO: make this resizing thing better
                    if img.height() > 24:
                        img = img.scaledToHeight(24)
                    if img.width() > 24:
                        img = img.scaledToWidth(24)

                    return img

    def sort(self, column_index, order=None) -> None:
        if self.columns:
            # TODO: instead of relying on column (positional) index we should
            #       be keeping a referencing to the column and improve 
            #       Column identity and equality comparison to restore sorting
            column = self.columns[column_index]
            self._last_sorting_column = column, column_index, order
            self._apply_sorting()

    def _apply_sorting(self):
        if self._last_sorting_column is not None:
            previous_column, column_index, order = self._last_sorting_column
            try:
                column = self.columns[column_index]
            except IndexError:
                # TODO: we should still try to find if the column is there in another index
                return
            else:
                if column.title != previous_column.title:
                    # not very likely to be the same column - avoid sorting incorrectly
                    return
            self.beginResetModel()
            def sort_key(item):
                value = column.get_value(item)
                if isinstance(value, float):
                    if math.isnan(value):
                        return float('-inf')
                return value
            self._filtered_items = sorted(self._filtered_items, key=sort_key, reverse=order)
            self.endResetModel()

    def setData(self, index: QModelIndex, value: Any, role: int) -> bool:
        if index.column() == 0 and role == Qt.CheckStateRole and self.checkable:
            item = self.items[index.row()]
            if value == Qt.Checked:
                self.checked_items.add(item)
                self.on_checked_items.emit(self.checked_items)
                return True
            elif value == Qt.Unchecked:
                self.checked_items.discard(item)
                self.on_checked_items.emit(self.checked_items)
                return True

        return False

    def headerData(self, section: int, orientation: Qt.Orientation, role: int) -> str:
        if orientation == Qt.Horizontal:
            if self.checkable:
                offset = 1

                # TEMP
                if section == 0:
                    return
                # END OF TEMP
            else:
                offset = 0

            if role == Qt.DisplayRole:
                column = self.columns[section - offset]  # O(1)
                if column.title is not None:
                    return column.title
            elif role == Qt.TextAlignmentRole:
                column = self.columns[section - offset]  # O(1)
                if len(self):
                    first_item = self.items[0]
                    align = column.get_align(first_item)
                else:
                    if (align := column.align) is AUTO_ALIGN:
                        align = Alignment.LEFT
                return to_qt_alignment(align)
            elif role == Qt.ForegroundRole:
                column = self.columns[section - offset]  # O(1)
                if column in self.filters:
                    return RED

            elif role == Qt.FontRole:
                column = self.columns[section - offset]  # O(1)
                if column in self.filters:
                    return self._font_bold

        return super().headerData(section, orientation, role)

    # QTableModel interface -------------------------------------------------------------------------------------------
    def get_column_by_index(self, index: int) -> Column:
        offset = 1 if self.checkable else 0
        return self.columns[max(index - offset, 0)]  # TODO: properly address checkable column here

    def get_item_by_index(self, index: int) -> Any:
        return self.items[index]

    def get_cell_value(self, row_index: int, column_index: int) -> Any:
        if column_index or not self.checkable:
            column = self.columns[column_index]  # O(1)
            item = self.items[row_index]  # O(1)
            return column.get_value(item)

    @property
    def items(self) -> List[Any]:
        """ read from the filtered items """
        return self._filtered_items

    @items.setter
    def items(self, items: Iterable[Any]) -> None:
        """ write to the original items (not filtered) """
        self._original_items = items
        self._apply_filters()
        try:
            self._apply_sorting()
        except (IndexError, KeyError) as ex:
            print('it was not possible to re-apply sorting configuration: {ex}')

    def set_filter(self, column: Column, expression: str) -> None:
        filter = Filter(column, expression)
        self.filters.add_filter(filter)
        self._apply_filters()

    def _apply_filters(self) -> None:
        self.beginResetModel()
        self.refresh_filtered_items()
        self.endResetModel()

    def clear_filters(self):
        self.filters.clear()
        self._apply_filters()

    def refresh_filtered_items(self) -> None:
        """
        the problem here is that when we are applying the filters we are
        creating a new list - can we do that better?
        instead we could keep the items inside a proxy object which
        holds an internal mapping for the indexes
        (index in the original items -> index in the filtered items)
        """
        if self.filters:
            self._filtered_items = list(self.filters.filter_items(self._original_items))  # TODO: we really want a list?
        else:
            self._filtered_items = self._original_items

    def _create_font(self, bold: bool = False) -> QFont:
        font = QFont(DEFAULT_FONT_NAME)
        font.setPixelSize(DEFAULT_FONT_SIZE_PX)
        if bold:
            font.setBold(True)
        return font


class DoubleClickContext:
    def __init__(self, index: QModelIndex, table: "QTable"):
        self.index = index
        self.__table = table

    @property
    def cell(self) -> Tuple[int, int]:
        return (self.row_index, self.column_index)

    @property
    def row_index(self) -> int:
        return self.index.row()

    @property
    def column_index(self) -> int:
        return self.index.column()

    @property
    def column(self) -> Column:
        model = self.__table.model()
        return model.get_column_by_index(self.column_index)

    @property
    def item(self) -> Any:
        model = self.__table.model()
        return model.get_item_by_index(self.row_index)

    @property
    def raw_value(self) -> Any:
        return self.column.get_value(self.item)

    @property
    def value(self) -> str:
        """ Returns the displayed value. """
        return self.column.get_displayed_value(self.item)  # TODO: should ask the column or should ask the model?


class SelectionContext:
    def __init__(self,
                 table: "QTable",
                 selected_indexes: List[QModelIndex],
                 current: QModelIndex,
                 added: Optional[List[QModelIndex]],
                 removed: Optional[List[QModelIndex]]):
        self.__table = weakref.ref(table)
        self.selected_model_indexes = selected_indexes
        self.added_model_indexes = added
        self.removed_model_indexes = removed
        self.current_model_index = current

    @property
    def selected_indexes(self) -> List[Cell]:
        return [(i.row(), i.column()) for i in self.selected_model_indexes]

    @property
    def added_indexes(self) -> Optional[List[Cell]]:
        if self.added_model_indexes:
            return [(i.row(), i.column()) for i in self.added_model_indexes]

    @property
    def removed_indexes(self) -> Optional[List[Cell]]:
        if self.removed_model_indexes:
            return [(i.row(), i.column()) for i in self.removed_model_indexes]

    @property
    def current_index(self) -> Cell:
        return (self.current_model_index.row(), self.current_model_index.column())

    # TODO: distinguish between when selection is based on single cell, single row,
    #       multiple cells, multiple rows
    @property
    def selected_items(self) -> List[Any]:
        table = self.__table()
        if table is not None:
            row_indexes = (i.row() for i in self.selected_model_indexes)
            return [table.model().get_item_by_index(i)
                    for i in set(row_indexes)]

    @property
    def selected_values(self) -> List[Any]:  # TODO: generator?
        table = self.__table()
        if table is not None:
            get_cell_value = table.model().get_cell_value
            for row_index, column_index in self.selected_indexes:
                yield((row_index, column_index), get_cell_value(row_index, column_index))


class QTable(QTableView):
    """
    A table has basically columns and a collection of items.
    """

    #: on_double_click signal is emitted when doubleClicked signal is handled
    on_double_click: Signal = Signal(DoubleClickContext)

    #: on_selection signal is emitted whenever the selection on the table changes
    on_selection: Signal = Signal(SelectionContext)

    #: signal to notify whenever the checked items change
    on_checked_items: Signal = Signal(set)

    def __init__(self,
                 columns: List[Column],
                 items: Optional[List[Any]] = None,
                 *,
                 checkable: bool = False,
                 context_menu: List["ContextMenuAction"] = None,
                 alternate_row_colors: bool = True,
                 checked_items: Collection[Any] = None,
                 sortable: bool = True,
                 parent: QObject = None,
                 convert_item = None,
                 ):
        super().__init__(parent=parent)
        self.columns = columns
        if items is None:
            items = []
        self.items = items
        self.context_menu = context_menu
        self.setAlternatingRowColors(alternate_row_colors)
        self.doubleClicked.connect(self.on_double_clicked)
        model = QTableModel(self.columns, self.items, checkable=checkable, checked_items=checked_items,
                            convert_item=convert_item)
        model.on_checked_items.connect(self.on_model_checked_items_changed)
        self.setModel(model)
        self.verticalHeader().setDefaultSectionSize(DEFAULT_ROW_HEIGHT)
        self.sortable = sortable
        self.__updating = False  # sentinel
        # TODO: improve the way we update the internals - maybe offering a high-level function that gets everything
        #       that is internal and is possible of updating?
        #       Also, need to eliminate the duplication here - we should only work in terms of what's inside the model
        #       this will prevent from view and model getting out of sync


        # Horizontal header with filter capabilities
        h_header = QFilterableHeaderView(Qt.Horizontal, parent=self)
        h_header.filterChanged.connect(self.on_filter_changed)
        h_header.setSectionsClickable(True)
        h_header.setSortIndicatorShown(True)
        self.setHorizontalHeader(h_header)

        self.__selection_mode_override = None

    def keyPressEvent(self, event: QEvent):
        """ Supports copying when multiple cells are selected. """
        if event.matches(QKeySequence.Copy):
            selected_indexes = self.selectedIndexes()
            if selected_indexes:
                data = ((i.row(), i.column(), self.model().data(i, role=Qt.DisplayRole)) for i in selected_indexes)
                rows = ((v for _, _, v in cells)
                        for _, cells in itertools.groupby(sorted(data), key=operator.itemgetter(0)))

                sio = StringIO()
                writer = csv.writer(sio, delimiter='\t')
                writer.writerows(rows)

                sio.seek(0)
                QApplication.clipboard().setText(sio.read())
        elif (event.modifiers() & Qt.ControlModifier) and (event.modifiers() & Qt.AltModifier):
            if self.__selection_mode_override is None:
                self.__selection_mode_override = self.__selection_mode
                self.set_selection_mode(SelectionMode.MULTI_CELLS)
                self.clearSelection()
        else:
            return super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QEvent) -> None:
        if not ((event.modifiers() & Qt.ControlModifier) and (event.modifiers() & Qt.AltModifier)):
            if self.__selection_mode_override is not None:
                self.set_selection_mode(self.__selection_mode_override)
                self.__selection_mode_override = None
        return super().keyReleaseEvent(event)

    def on_filter_changed(self, column: Column, expression: str) -> None:
        self.model().set_filter(column, expression)

    def clear_filters(self):
        self.model().clear_filters()

    @property
    def columns(self) -> List[Column]:
        return self._columns

    @columns.setter
    def columns(self, columns: List[Column]):
        self._columns = columns
        # refresh the model
        if self.model() is not None:
            self.model().columns = columns

    @property
    def items(self) -> List:
        return self._items

    @items.setter
    def items(self, items: List[Any]):
        self._items = items
        # refresh the model
        if self.model() is not None:
            self.model().items = items

    @property
    def checkable(self) -> bool:
        return self.model().checkable

    @checkable.setter
    def checkable(self, checkable: bool):
        self.model().checkable = checkable

    @property
    def sortable(self) -> bool:
        return self.isSortingEnabled()

    @sortable.setter
    def sortable(self, sortable: bool) -> None:
        self.setSortingEnabled(sortable)

    @contextlib.contextmanager
    def updating_internals(self):
        self.__updating = True

        try:
            self.model().beginResetModel()
            yield
        finally:
            self.__updating = False
            self.model().endResetModel()
            if not hasattr(self, '_adjusted_columns_size') and self.columns and self.items:
                # print('adjusting column sizes for the first and only time')
                self.adjust_column_sizes()
                self._adjusted_columns_size = True

    def refresh(self):
        m = self.model()
        m.refresh_filtered_items()  # TODO: find a better way without requiring this
        top_left = m.index(0, 0)
        bottom_right = m.index(len(self.items), len(self.columns))
        m.dataChanged.emit(top_left, bottom_right)

    def refresh_one_cell(self, row: int, col: int) -> None:
        # how about filtering and sorting here?
        # perhaps we can say that ticking tables cannot be filtered or sorted?
        m = self.model()
        index = m.index(row, col)
        index2 = m.index(row, col+1)
        m.dataChanged.emit(index, index2)

    def set_selection_mode(self, selection_mode: SelectionMode):
        if selection_mode == SelectionMode.SINGLE_CELL:
            self.setSelectionMode(self.SingleSelection)
            self.setSelectionBehavior(self.SelectItems)
        elif selection_mode == SelectionMode.MULTI_CELLS:
            self.setSelectionMode(self.ExtendedSelection)
            self.setSelectionBehavior(self.SelectItems)
        elif selection_mode == SelectionMode.SINGLE_ROW:
            self.setSelectionMode(self.SingleSelection)
            self.setSelectionBehavior(self.SelectRows)
        elif selection_mode == SelectionMode.MULTI_ROWS:
            self.setSelectionMode(self.ExtendedSelection)
            self.setSelectionBehavior(self.SelectRows)
        self.__selection_mode = selection_mode

    # Header visibility controls

    def show_vertical_header(self):
        self.verticalHeader().show()

    def hide_vertical_header(self):
        self.verticalHeader().hide()

    def set_vertical_header_visible(self, visible: bool):
        self.verticalHeader().setVisible(visible)

    def show_horizontal_header(self):
        self.horizontalHeader().show()

    def hide_horizontal_header(self):
        self.horizontalHeader().hide()

    def set_horizontal_header_visible(self, visible: bool):
        self.horizontalHeader().setVisible(visible)

    # Context Menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        default_actions = [ResetFiltersAction(table=self)]
        context_menu = self.context_menu + default_actions

        convert_item = self.model().convert_item
        context = MenuActionContext(pos=event.pos(), table=self, convert_item=convert_item)
        enabled_actions = [a for a in context_menu if a.is_enabled(context)]
        if enabled_actions:
            menu = QMenu(parent=self)
            for context_menu_action in enabled_actions:
                action = QAction(parent=menu)
                action.setText(context_menu_action.get_caption(context))

                def on_action_triggered(checked: bool = False, *, context_menu_action=context_menu_action):
                    return context_menu_action.execute(context)

                action.triggered.connect(on_action_triggered)
                menu.addAction(action)

            pos = self.mapToGlobal(event.pos())
            menu.exec_(pos)

    # Double Click

    def on_double_clicked(self, index: QModelIndex):
        """
        Function handler listening whenever Qt fires the doubleClicked event.
        """
        double_click_context = DoubleClickContext(index=index, table=self)
        self.on_double_click.emit(double_click_context)

    # Selection

    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection):
        super().selectionChanged(selected, deselected)
        selection_context = SelectionContext(
            table=self,
            selected_indexes=self.selectionModel().selectedIndexes(),
            added=selected.indexes(),
            removed=selected.indexes(),
            current=self.currentIndex(),
        )
        self.on_selection.emit(selection_context)

    def get_current_selection_context(self) -> SelectionContext:
        selection_context = SelectionContext(
            table=self,
            selected_indexes=self.selectionModel().selectedIndexes(),
            added=None,
            removed=None,
            current=self.currentIndex(),
        )
        return selection_context

    # Checked Items

    def on_model_checked_items_changed(self, checked_items: Set):
        """ Just passes along the notification of the change on the checked items.
        """
        self.on_checked_items.emit(checked_items)

    @property
    def checked_items(self) -> Set:
        if self.checkable:
            return self.model().checked_items
        else:
            return set()

    # Column Sizes
    def adjust_column_sizes(self) -> None:
        fm = self.fontMetrics()
        model = self.model()
        items = model.items
        for i, col in enumerate(self.columns):
            if col.size == 'auto':
                # THIS IS SUPER SLOW!
                self.resizeColumnToContents(i)
            elif col.size == 'just':
                # Consider the font metrics of the view
                # Here we are taking into account only the first thousand rows
                # because otherwise this would be very slow for millions of rows
                # It does not take into consideration the text in the header
                get_displayed_value = col.get_displayed_value
                size = 20
                for row_index in range(min(model.rowCount(), 1000)):
                    size = max(size, fm.width(get_displayed_value(items[row_index])))
                self.horizontalHeader().resizeSection(i, size + 10)
            elif isinstance(col.size, int):
                self.horizontalHeader().resizeSection(i, col.size)
            elif col.size == 'ignore':
                continue
            else:
                raise ValueError(f'Invalid column size: {col.size}')


class MenuActionContext:
    """
    This class represents what gets passed to the context menu actions (ContextMenuAction)
    when the user requests a context menu on the table (for example, by performing a right-click)
    """

    def __init__(self,
                 pos: QPoint,
                 table: QTable,
                 convert_item = default_convert_item,
                 ):
        self.__table = table
        self.pos = pos
        if convert_item is None:
            convert_item = default_convert_item
        self.convert_item = convert_item

    @property
    def model(self) -> QTableModel:
        return self.__table.model()

    @property
    def index(self) -> QModelIndex:
        return self.__table.indexAt(self.pos)

    @property
    def row_index(self) -> int:
        return self.index.row()

    @property
    def column_index(self) -> int:
        return self.index.column()

    @property
    def item(self) -> Any:
        return self.convert_item(self.raw_item)

    @property
    def raw_item(self) -> Any:
        if (row_index := self.row_index) != -1:
            return self.model.get_item_by_index(row_index)

    @property
    def column(self) -> Column:
        if (col_index := self.column_index) != -1:
            return self.model.get_column_by_index(col_index)

    @property
    def raw_value(self) -> Any:
        if (item := self.item) is not None:  # TODO: use a different sentinel here?
            return self.column.get_value(item)

    @property
    def value(self) -> str:
        """ Returns the displayed value. """
        if (item := self.item) is not None:  # TODO: use a different sentinel here?
            return self.column.get_displayed_value(item)


class ContextMenuAction(ABC):
    @abstractmethod
    def is_enabled(self, context: MenuActionContext) -> bool:
        pass

    @abstractmethod
    def get_caption(self, context: MenuActionContext) -> bool:
        pass

    @abstractmethod
    def execute(self, context: MenuActionContext):
        pass


class ResetFiltersAction(ContextMenuAction):
    def __init__(self, table: QTable):
        self.table = table

    def is_enabled(self, context: MenuActionContext) -> bool:
        return bool(self.table.model().filters)

    def get_caption(self, context: MenuActionContext) -> bool:
        return 'Reset Filters'

    def execute(self, context: MenuActionContext) -> None:
        self.table.clear_filters()


from qtpy.QtWidgets import QWidget, QHeaderView, QLineEdit, QPushButton, QGroupBox, QHBoxLayout, QVBoxLayout, QLabel


class QFilterWidget(QWidget):
    def __init__(self, column, callback, filters, *, parent):
        super().__init__(parent)
        self.column = column
        self.callback = callback
        self.filters = filters
        self._setup_ui()

    def _setup_ui(self):
        caption = f'Filter: {self.column.title}'
        self.setWindowTitle(caption)

        layout = QVBoxLayout()

        label = QLabel(caption)
        label.font().setBold(True)
        layout.addWidget(label)

        self._input_field = input_field = QLineEdit()
        layout.addWidget(input_field)
        input_field.returnPressed.connect(self._notify_callback)

        if (current_filter := self.filters.get(self.column)) is not None:
            input_field.setText(str(current_filter))

        gb = QGroupBox()
        gb.setFlat(True)
        gb.setTitle('')

        gb_layout = QHBoxLayout()
        btn_ok = QPushButton('OK')
        btn_ok.setDefault(True)
        btn_ok.clicked.connect(self._notify_callback)

        btn_cancel = QPushButton('Cancel')
        btn_cancel.setDefault(False)
        btn_cancel.clicked.connect(self.close)

        gb_layout.addStretch()
        gb_layout.addWidget(btn_ok)
        gb_layout.addWidget(btn_cancel)

        gb.setLayout(gb_layout)

        layout.addWidget(gb)

        self.setLayout(layout)

        self.setWindowFlags(Qt.Popup)
        self.resize(200, 80)

    def show(self, global_pos):
        self.move(global_pos)
        super().show()
        self._input_field.setFocus()

    def _notify_callback(self, *args):
        expression = str(self._input_field.text()).strip()
        self.callback(self.column, expression)
        self._input_field = None
        self.close()


class QFilterableHeaderView(QHeaderView):
    filterChanged = Signal(Column, str)

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            table: QTable = self.parent()
            model: QTableModel = table.model()
            column_index = self.logicalIndexAt(event.pos())
            column = model.get_column_by_index(column_index)
            filter_widget = QFilterWidget(column, self.filter_callback, model.filters, parent=table)
            filter_widget.show(event.globalPos())

        return super().mousePressEvent(event)

    def filter_callback(self, column: Column, expression: str):
        # TODO: use column index?
        self.filterChanged.emit(column, expression)


def debug_trace():
    """
    Set a tracepoint in the Python debugger that works with Qt
    """
    import pdb
    import sys
    from PyQt5.QtCore import pyqtRemoveInputHook
    pyqtRemoveInputHook()
    # set up the debugger
    debugger = pdb.Pdb()
    debugger.reset()
    # custom next to get outside of function scope
    debugger.do_next(None)  # run the next command
    users_frame = sys._getframe().f_back  # frame where the user invoked `pyqt_set_trace()`
    debugger.interaction(users_frame, None)

    # TODO: should call QtCore.pyqtRestoreInputHook() ?


if __name__ == '__main__':
    import uuid

    # Making Ctrl+C work
    import signal

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Fixing PyQt issue on MacOS BigSur
    import os

    os.environ["QT_MAC_WANTS_LAYER"] = "1"

    app = QApplication([])


    def tooltip_callback(table_context: TableContext) -> str:
        item = table_context.item
        return f"{item.name} is {item.age} years old"


    def age_cell_style(table_context: TableContext) -> CellStyle:
        if table_context.item.age > 30:
            return CellStyle(color=QColor(200, 20, 100))


    columns = [
        Column("name", title="Name"),
        Column("age", align=Alignment.RIGHT, tooltip=tooltip_callback,
               cell_style=age_cell_style)
    ]


    @dataclass(frozen=True)
    class Person:
        id: str
        name: str
        age: int
        sex: str


    people_kwargs = [
        dict(name="John", age=33, sex="M"),
        dict(name="Pam", age=22, sex="F"),
    ]

    items = [Person(id=uuid.uuid4(), **kw) for kw in itertools.islice(itertools.cycle(people_kwargs), 0, 10)]


    def echo_factory(n: int) -> ContextMenuAction:
        class EchoAction(ContextMenuAction):
            def is_enabled(self, context):
                return True

            def get_caption(self, context):
                return f"Echo {n}"

            def execute(self, context):
                print(f"Echo {n} - Cell: ({context.row_index}, "
                      f"{context.column_index}) has value: {context.raw_value!r}")

        return EchoAction()


    context_menu_actions = [
        *(echo_factory(n) for n in range(1, 11)),
    ]

    table = QTable(columns, items, checkable=True, context_menu=context_menu_actions)
    table.set_selection_mode(SelectionMode.MULTI_CELLS)
    # table.set_selection_mode(SelectionMode.SINGLE_ROW)
    table.set_vertical_header_visible(False)
    table.show_horizontal_header()


    # table.setStyleSheet("background-color: rgb(255, 255, 245); alternate-background-color: rgb(220,200,220);")

    def double_click_callback(context: DoubleClickContext):
        print(f"Double click happened:",
              f"  cell = {context.cell}",
              f"  column = {context.column}",
              f"  raw_value = {context.raw_value}",
              sep="\n")


    table.on_double_click.connect(double_click_callback)


    def selection_callback(context: SelectionContext):
        print(
            f"Selected indexes: {context.selected_indexes}",
            f"Added: {context.added_indexes}",
            f"Removed: {context.removed_indexes}",
            f"Current: {context.current_index}",
            sep="\n",
        )


    table.on_selection.connect(selection_callback)

    def checked_items_callback(checked_items: set):
        print("Checked items:", checked_items)

    table.on_checked_items.connect(checked_items_callback)


    table.show()

    app.exec_()



import datetime
from enum import Enum
from numbers import Number
from operator import itemgetter
from typing import Union, Callable, Optional, Any, Sequence, Mapping, List, Dict, Container

from qtpy.QtCore import Qt
from qtpy.QtGui import QColor

from enamlext.qt.qt_dataframe import DataFrameProxy
from enamlext.qt.table.defs import CellStyle, ColumnSize
from enamlext.qt.table.table_context import TableContext


class Alignment(str, Enum):
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


AUTO_ALIGN = object()  # sentinel


class Column:
    def __init__(self,
                 key: Union[str, Callable],
                 title: Optional[str] = None,
                 align: Alignment = AUTO_ALIGN,
                 tooltip: Optional[Union[str, Callable]] = None,
                 # TODO: how to specify the types for the signature of the callback here?
                 cell_style: Optional[Callable] = None,
                 use_getitem: bool = False,
                 fmt: str = '',
                 size: Union[ColumnSize, int] = ColumnSize.AUTO,
                 image: Optional[str] = None, # TODO: support callback
                 collect_stats: bool = False,
                 ):
        self.key = key
        self._link_get_value_method(key, use_getitem)
        if title is None and isinstance(key, str):
            title = key.title()
        self.title = title
        self.align = align
        self.tooltip = tooltip
        self.cell_style = cell_style
        self.fmt = fmt
        self.size = size  # TODO: should we also support callable?
        self.image = image
        if cell_style is not None:
            self.get_cell_style = cell_style
        self.collect_stats = collect_stats
        self.stats = {}

    def record_stats(self, role: int, elapsed: float):
        if role not in self.stats:
            self.stats[role] = {'n_calls': 0, 'cum_time': 0, 'avg_time': 0}

        stats_for_role = self.stats[role]
        stats_for_role['n_calls'] += 1
        stats_for_role['cum_time'] += elapsed
        stats_for_role['avg_time'] = stats_for_role['cum_time'] / stats_for_role['n_calls']

    def total_cost(self):
        if not self.stats:
            return 0

        return sum((stats['cum_time'] for stats in self.stats))

    def _link_get_value_method(self, key, use_getitem):
        if callable(key):
            self.get_value = key  # we re-wire the get_value() here
        elif use_getitem:
            self.get_value = self.get_value_by_getitem_lookup
        else:
            self.get_value = self.get_value_by_attribute_lookup

    def get_value(self, item: Any) -> Any:
        raise RuntimeError('get_value() not dynamic linked during initialization')

    def get_value_by_attribute_lookup(self, item: Any) -> Any:
        return getattr(item, self.key)

    def get_value_by_getitem_lookup(self, item: Any) -> Any:
        return item[self.key]
        # TODO: dataframe will always need to be by index (not by name)

    def get_displayed_value(self, item: Any) -> str:
        value = self.get_value(item)
        if callable(self.fmt):
            return self.fmt(value)
        else:
            return format(value, self.fmt) if value is not None else ''

    def get_tooltip(self, table_context: "TableContext") -> str:
        if self.tooltip is not None:
            if callable(self.tooltip):
                return str(self.tooltip(table_context))  # TODO: is it a good practice to enforce str() here?
            else:
                return self.tooltip  # str
        else:
            return repr(table_context.raw_value)

    def get_cell_style(self, table_context: "TableContext") -> Optional[CellStyle]:
        if self.cell_style is not None:
            return self.cell_style(table_context) or CellStyle()

    def get_align(self, item: Any) -> Alignment:
        if self.align is AUTO_ALIGN:
            # TODO: consider shortcircuiting this - maybe
            #       maybe this should only be done once, for the first time - then shortcircuit
            #       or maybe the user wants to have mixed alignments on the same column, so it can provide a callback
            value = self.get_value(item)
            align = self.resolve_column_alignment_based_on_value(value)
            return align
        else:
            return self.align

    def resolve_column_alignment_based_on_value(self, value: Any) -> Alignment:
        if isinstance(value, Number):
            return Alignment.RIGHT
        elif isinstance(value, (datetime.datetime, datetime.date)):
            return Alignment.CENTER
        elif isinstance(value, str):
            return Alignment.LEFT
        else:
            return Alignment.LEFT

    def get_image(self, table_context: 'TableContext') -> Optional[str]:  # TODO: should return Path?
        if self.image:
            return str(self.image)


# Auto-Generation of Columns

def make_title(title: str):
    return str(title).replace('_', ' ').title()


def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr((T := type(obj)), '_fields') and hasattr(T, '_asdict')


RED = QColor(Qt.red)
NEGATIVE_NUMBER_CELL_STYLE = CellStyle(color=RED)


def get_cell_style_for_negative_numbers(table_context: TableContext) -> CellStyle:
    if table_context.raw_value < 0:
        return NEGATIVE_NUMBER_CELL_STYLE


def generate_columns(items: Sequence, *, hints: Optional[Dict] = None,
                     include: Optional[Container[str]] = None,
                     exclude: Optional[Container[str]] = None) -> List[Column]:
    """
    hints: only make sense with named keys?
           hints are a dict of column id -> kwargs dict that will
           get passed to the Column object

    Args:
        include: list of column keys/ids that will be included. The order
            of the generated columns will be respected according to the
            include values. Use this to ensure the order of the columns.

        exclude: list of column keys/ids to be excluded. The order
            of the generated columns will be according to the order of
            appearance in the first item.

    Note: you cannot use include and exclude at the same time. An attempt
        to do so will result in raising a ValueError exception.
    """
    if hints is None:
        hints = {}
    if include and exclude:
        raise ValueError('Cannot use include and exclude simultaneously.')
    if exclude is not None:
        exclude_set = set(exclude)
    if include is not None:
        include_set = set(include)
    if not len(items):
        return []
    first_row = items[0]
    columns = {}  # column_index -> Column

    if isinstance(items, DataFrameProxy):
        import pandas as pd
        import numpy as np
        df = items.df
        for i, (name, dtype) in enumerate(df.dtypes.items()):
            if exclude is not None and name in exclude_set:
                continue
            if include is not None and name not in include_set:
                continue
            if not isinstance(dtype, pd.core.dtypes.dtypes.CategoricalDtype) and np.issubdtype(dtype, np.number):
                align = Alignment.RIGHT
                style = get_cell_style_for_negative_numbers
            else:
                align = Alignment.LEFT
                style = None

            kwargs = {'title': make_title(name), 'align': align, 'cell_style': style,
                      'use_getitem': True}

            hint = hints.get(name, {})
            kwargs.update(hint)

            column = Column(itemgetter(i), **kwargs)

            if include is not None:
                column_index = include.index(name)
            else:
                column_index = i

            column.df_index = i

            columns[column_index] = column

    elif isinstance(first_row, tuple):
        if is_namedtuple(first_row):
            fields = type(first_row)._fields
        else:
            fields = None
        for i, value in enumerate(first_row):
            field_name = fields[i] if fields is not None else object()
            if exclude is not None and (set([i, field_name]) & exclude_set):
                continue
            if include is not None and not (set([i, field_name]) & include_set):
                continue
            if fields is not None:
                title = make_title(fields[i])
            else:
                title = str(i)
            kwargs = {'title': title}
            if isinstance(value, Number):
                kwargs['align'] = Alignment.RIGHT
            else:
                kwargs['align'] = Alignment.LEFT
            hint = hints.get(i, {})
            kwargs.update(hint)
            column = Column(itemgetter(i), **kwargs)
            if include is not None:
                for v in (i, field_name):
                    try:
                        column_index = include.index(v)
                    except ValueError:
                        continue
                    else:
                        break
            else:
                column_index = i
            columns[column_index] = column

    elif isinstance(first_row, Mapping):
        for i, (key, value) in enumerate(first_row.items()):
            if exclude is not None and key in exclude_set:
                continue
            if include is not None and key not in include_set:
                continue
            if isinstance(key, str):
                title = make_title(key)
            else:
                title = str(key)

            kwargs = {'title': title}
            if isinstance(value, Number):
                kwargs['align'] = Alignment.RIGHT
            else:
                kwargs['align'] = Alignment.LEFT

            hint = hints.get(key, {})
            kwargs.update(hint)
            column = Column(itemgetter(key), **kwargs)

            if include is not None:
                column_index = include.index(key)
            else:
                column_index = i
            columns[column_index] = column

    elif hasattr(type(first_row), '__dataclass_fields__'):
        fields = type(first_row).__dataclass_fields__
        for i, field in enumerate(fields):
            if exclude is not None and field in exclude_set:
                continue
            if include is not None and field not in include_set:
                continue
            value = getattr(first_row, field)
            title = make_title(field)
            kwargs = {'title': title}
            if isinstance(value, Number):
                kwargs['align'] = Alignment.RIGHT
            else:
                kwargs['align'] = Alignment.LEFT

            hint = hints.get(field, {})
            kwargs.update(hint)

            column = Column(field, **kwargs)
            if include is not None:
                column_index = include.index(field)
            else:
                column_index = i
            columns[column_index] = column

    ordered_columns = [columns[c] for c in sorted(columns)]
    return ordered_columns



from typing import Iterable, Generator, Any, List, Optional

from enamlext.qt.table.column import Column


class Filter:
    """ One filter, bound to one column.
    """
    def __init__(self, column: Column, expression: str):
        self.column = column
        self.expression = expression.strip()  # as entered by the user
        self._evaluate_filter = self._generate_filter_evaluation_callback()

    def __str__(self):
        return self.expression

    def _expression_startswith_operator(self, expression: str) -> bool:
        operators = {'>', '<', '>=', '<=', '==', '!='}

        for op in operators:
            if expression.startswith(op):
                return True

        return False

    def _generate_filter_evaluation_callback(self):
        if self._expression_startswith_operator(self.expression):
            def callback(x):
                ns = {'x': x}
                try:
                    return eval(f'x {self.expression}', ns)
                except (NameError, SyntaxError):
                    return False

        else:
            def callback(x):
                return self.expression.lower() in str(x).lower()

        return callback

    def __call__(self, item: Any) -> bool:
        value = self.column.get_value(item)
        return self._evaluate_filter(value)

# TODO: IDEA
#       we are using the Column's get_value to get the value
#       that is then used to evaluate the filter expression
#       We can have that as default behaviour, but we could also
#       allow for the underlying model to apply their own filters
#       This way it would make possible to use, for example,
#       pandas DataFrame's own filtering capabilities more
#       efficiently.


class TableFilters:
    """ Collection of Filters.
    """
    def __init__(self, filters: List[Filter] = None):
        if filters is None:
            filters = {}
        self.filters = {f.column: f for f in filters}

    def __len__(self):
        return len(self.filters)

    def __contains__(self, column):
        return column in self.filters

    def filter_items(self, items: Iterable) -> Generator:
        for item in items:
            if self.filter(item):
                yield item

    def filter(self, item: Any) -> bool:
        """ Returns True if the given item is included after evaluating all the filters.
        """
        for filter in self.filters.values():
            if not filter(item):
                return False
        return True

    def add_filter(self, filter: Filter) -> None:
        if filter.expression:
            self.filters[filter.column] = filter
        else:
            self.filters.pop(filter.column, None)

    def get(self, column: Column) -> Optional[Filter]:
        return self.filters.get(column)

    def clear(self):
        self.filters = {}



from dataclasses import dataclass
from typing import Optional

from enamlext.qt.qtable import SelectionContext


@dataclass
class TableSelectionSummary:
    sum: float
    count_numbers: int
    min: float
    max: float
    values: list

    @classmethod
    def from_selection_context(cls, selection_context: SelectionContext) -> "TableSelectionSummary":
        return compute_summary([v for _, v in selection_context.selected_values])

    @property
    def count(self) -> int:
        return len(self.values)

    @property
    def avg(self) -> Optional[float]:
        if self.count_numbers:
            return self.sum / self.count_numbers

    @property
    def diff(self) -> Optional[float]:
        """ Returns the absolute difference/distance between 2 selected numbers
        when all is selected are those 2 numbers and nothing else.
        """
        if self.count == self.count_numbers == 2:
            v1, v2 = self.values
            try:
                return abs(v1 - v2)
            except Exception:
                return None

    def __str__(self) -> str:
        parts = [f'Count: {self.count}']
        if self.count_numbers:
            parts += [
                f'Average: {self.avg}',
                f'Sum: {self.sum}',
                f'CountNumbers: {self.count_numbers}',
                f'Min: {self.min}',
                f'Max: {self.max}',
            ]
        if (diff := self.diff) is not None:
            parts.append(f'Diff: {diff}')

        return '   '.join(parts)


# TODO: Qt selection model has a better/more clever way to compute the summary
#       based on what's been deselected and selected since "last time"
def compute_summary(selected_values: list) -> TableSelectionSummary:
    sum_ = 0.0
    count_numbers = 0
    min_value = float('inf')
    max_value = -float('inf')

    for value in selected_values:
        try:
            sum_ += value
        except TypeError:
            pass
        else:
            count_numbers += 1
            min_value = min(min_value, value)
            max_value = max(max_value, value)

    return TableSelectionSummary(sum=sum_, count_numbers=count_numbers,
                                 min=min_value, max=max_value, values=selected_values)


from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QPushButton, QApplication, QHBoxLayout, QLayout
from PyQt5.QtGui import QPixmap, QColor, QPainter, QBrush, QPen
from PyQt5.QtCore import Qt
from atom.api import Typed
from enaml.qt.qt_control import QtControl

from enamlext.widgets.traffic_light import ProxyTrafficLight


class TrafficLightWidget(QWidget):
    """
    A custom PyQt widget that represents a traffic light with red, yellow, and green lights.
    """

    def __init__(self, radius: int = 20, orientation: str = 'vertical', parent: QWidget | None = None):
        super().__init__(parent=parent)

        # Set the background color to black
        self.setStyleSheet("background-color: black;")

        # Set the radius of the circles
        self.radius = radius

        # Initialize the layout based on orientation
        if orientation == 'vertical':
            layout = QVBoxLayout()
            self.setFixedSize(self.radius * 2 + 20, self.radius * 6 + 20)
        else:
            layout = QHBoxLayout()
            self.setFixedSize(self.radius * 6 + 20, self.radius * 2 + 20)

        layout.setSpacing(0)  # Remove spacing between QLabel widgets

        # Create QLabel widgets for each light
        self.red_light = QLabel()
        self.yellow_light = QLabel()
        self.green_light = QLabel()

        # Set QLabel background to black
        for light in [self.red_light, self.yellow_light, self.green_light]:
            light.setStyleSheet("background-color: black;")

        # Add lights to layout
        layout.addWidget(self.red_light)
        layout.addWidget(self.yellow_light)
        layout.addWidget(self.green_light)

        # Set the layout
        self.setLayout(layout)

        # Turn off all lights initially
        self.set_light(None)

    def set_radius(self, new_radius: int) -> None:
        """
        Dynamically set the radius of the lights and redraw the widget.

        :param new_radius: The new radius for the lights.
        """
        self.radius = new_radius
        if isinstance(self.layout(), QVBoxLayout):
            self.setFixedSize(self.radius * 2 + 20, self.radius * 6 + 20)
        else:
            self.setFixedSize(self.radius * 6 + 20, self.radius * 2 + 20)
        self.set_light(None)  # Force a redraw with the new radius
        self.update()  # Force a repaint of the widget

    def set_light(self, color: str) -> None:
        """
        Set the light to the given color ('red', 'yellow', 'green') or turn off all lights (None).

        :param color: The color of the light to turn on, or None to turn off all lights.
        """
        # Define the off and on colors
        off_color = QColor(128, 128, 128)
        on_colors = {
            'red': QColor(255, 0, 0),
            'yellow': QColor(255, 255, 0),
            'green': QColor(0, 255, 0)
        }

        # Turn off all lights
        self._set_light_color(self.red_light, off_color)
        self._set_light_color(self.yellow_light, off_color)
        self._set_light_color(self.green_light, off_color)

        # Turn on the selected light
        if color in on_colors:
            selected_light = getattr(self, f"{color}_light")
            self._set_light_color(selected_light, on_colors[color])

    def _set_light_color(self, label: QLabel, color: QColor) -> None:
        """
        Set the background color of a QLabel to represent a light.

        :param label: The QLabel to modify.
        :param color: The QColor to set.
        """
        pixmap = QPixmap(self.radius * 2, self.radius * 2)
        pixmap.fill(Qt.transparent)
        painter = QPainter(pixmap)
        painter.setBrush(QBrush(color))
        painter.setPen(QPen(Qt.black, 2))
        painter.drawEllipse(0, 0, self.radius * 2, self.radius * 2)
        painter.end()

        label.setPixmap(pixmap)
        label.setAlignment(Qt.AlignCenter)


class QtTrafficLight(QtControl, ProxyTrafficLight):
    """ A Qt implementation of an Enaml ProxyTrafficLight.
    """
    #: A reference to the widget created by the proxy.
    widget = Typed(TrafficLightWidget)

    # --------------------------------------------------------------------------
    # Initialization API
    # --------------------------------------------------------------------------
    def create_widget(self):
        """ Create the underlying TrafficLightWidget object.
        """
        d = self.declaration
        self.widget = TrafficLightWidget(radius=d.radius, parent=self.parent_widget())

    def init_widget(self):
        """ Initialize the underlying widget.
        """
        super().init_widget()
        d = self.declaration
        self.set_color(d.color)

    # --------------------------------------------------------------------------
    # ProxyTrafficLight API
    # --------------------------------------------------------------------------
    def set_color(self, color: str | None) -> None:
        self.widget.set_light(color)

    def set_radius(self, radius: int) -> None:
        self.widget.set_radius(radius)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    main_widget = QWidget()
    main_layout = QVBoxLayout()

    widget = TrafficLightWidget(radius=10, orientation='horizontal')
    main_layout.addWidget(widget)

    button_layout = QHBoxLayout()

    for color in ['red', 'yellow', 'green', None]:
        button = QPushButton(f"Turn {color}" if color else "Turn off")
        button.clicked.connect(lambda checked, color=color: widget.set_light(color))
        button_layout.addWidget(button)

    main_layout.addLayout(button_layout)
    main_widget.setLayout(main_layout)
    main_widget.show()

    sys.exit(app.exec_())



from dataclasses import dataclass
from typing import Any

import pytest

from enamlext.qt.qtable import QTable, Qt, QModelIndex
from enamlext.qt.table.column import Column, Alignment


class QTestTable(QTable):
    """
    A QTable with some funky testing capabilities.
    """

    def index(self, row_index: int, column_index: int) -> QModelIndex:
        return self.model().index(row_index, column_index)

    def data(self, row_index: int, column_index: int, role: int) -> Any:
        index = self.index(row_index, column_index)
        return self.model().data(index, role)

    def text(self, row_index: int, column_index: int) -> str:
        return self.data(row_index, column_index, Qt.DisplayRole)


@pytest.fixture
def table(qtbot):
    table = QTestTable(columns=[])
    qtbot.addWidget(table)
    yield table


def test_displayed_data_basic(table):
    @dataclass
    class Person:
        name: str
        age: int

    items = [
        Person(name="John", age=33),
        Person(name="Pam", age=22),
    ]

    columns = [
        Column("name", title="Name"),
        Column("age", align=Alignment.RIGHT),
    ]

    with table.updating_internals():
        table.columns = columns
        table.items = items

    assert table.text(1, 0) == "Pam"



from atom.api import Typed, Int
from enaml.qt.QtWidgets import QFontComboBox
from enaml.qt.QtGui import QFont
from enaml.qt.qt_control import QtControl

from enamlext.widgets.font_combo_box import ProxyFontComboBox


# cyclic notification guard flags
INDEX_GUARD = 0x1


class QtFontComboBox(QtControl, ProxyFontComboBox):
    """ A Qt implementation of an Enaml FontComboBox.
    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QFontComboBox)

    # TODO: context manager
    #: Cyclic notification guard. This a bitfield of multiple guards.
    _guard = Int(0)

    # Initialization API
    def create_widget(self):
        """ Create the QFontComboBox widget.
        """
        self.widget = QFontComboBox(self.parent_widget())

    def init_widget(self):
        """ Create and initialize the underlying widget.
        """
        super().init_widget()
        self.widget.currentFontChanged.connect(self.on_font_changed)

    # Signal Handlers
    def on_font_changed(self, qfont: QFont):
        """ The signal handler for the font changed signal.
        """
        if not self._guard & INDEX_GUARD:
            self.declaration.font = qfont.family()

    def set_font(self, font: str):
        self._guard |= INDEX_GUARD
        try:
            qfont = QFont()
            qfont.fromString(font)
            self.widget.setCurrentFont(qfont)
        finally:
            self._guard &= ~INDEX_GUARD



import logging
import threading
from typing import Callable

import time
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def _monitor_df_changes(df_proxy):
    if df_proxy.instrumentation_enabled:
        logger.info('Starting DataFrameProxy thread for monitoring changes and refreshing ticking cells '
                    f'{threading.current_thread()}')
    
    from enaml.application import deferred_call

    current_values = df_proxy.values
    interval = df_proxy.tick_interval_ms / 1_000

    while df_proxy.is_active():
        time.sleep(interval)

        if df_proxy.instrumentation_enabled:
            t0 = time.perf_counter()

        new_values = df_proxy.df.values.copy()

        row_indexes, col_indexes = np.where(~((current_values == new_values) | (pd.isna(current_values) & pd.isna(new_values))))

        if len(row_indexes):
            deferred_call(df_proxy.update_values_and_refresh_cells, new_values, row_indexes, col_indexes)

        current_values = new_values

        if df_proxy.instrumentation_enabled:
            elapsed = time.perf_counter() - t0
            logger.info(f'It took {t1 -t0:.3f} s inside the monitoring thread')

    if df_proxy.instrumentation_enabled:
        logger.info(f'Thread died: DataFrameProxy ticking monitor {threading.current_thread()}')


class DataFrameProxy:
    def __init__(self,
                 df: pd.DataFrame,
                 *,
                 tick_interval_ms: int = 0,
                 refresh_cells_callback: Callable[[list, list], None] | None = None,
                 instrumentation_enabled: bool = False):
        if tick_interval_ms > 0 and refresh_cells_callback is None:
            raise ValueError('You must pass a callback to handle refreshing cells when you pass '
                             'a tick_interval_ms > 0.')
        if tick_interval_ms <= 0 and refresh_cells_callback is not None:
            raise ValueError('You should specify a tick_interval_ms refresh interval in miliseconds when you '
                             'pass a refresh_cells_callback.')
        self.values = df.values
        self.df = df
        self.tick_interval_ms = tick_interval_ms
        self.refresh_cells_callback = refresh_cells_callback
        self.instrumentation_enabled = instrumentation_enabled
        self._is_active = True
        if self.is_ticking:
            import threading
            t = threading.Thread(target=_monitor_df_changes,
                                 args=(self,),
                                 daemon=True)
            t.start()

    def update_values_and_refresh_cells(self, new_values, row_indexes, col_indexes):
        # must be called in the main thread!
        if self.instrumentation_enabled:
            t0 = time.perf_counter()

        self.values = new_values
        self.refresh_cells_callback(row_indexes, col_indexes)

        if self.instrumentation_enabled:
            elapsed = time.perf_counter() - t0
            changed_col_names = {i: self.df.columns[i] for i in set(col_indexes)}
            logger.info(f'It took {elapsed:.3f} s inside the main thread refreshing {len(row_indexes)} cells - modified cols: {changed_col_names}')

    @property
    def is_ticking(self):
        return self.tick_interval_ms > 0

    def deactivate(self):
        self._is_active = False

    def is_active(self):
        return self._is_active

    def __getitem__(self, item):
        return self.values[item]

    def __len__(self):
        return len(self.values)



import functools
import threading

from atom.api import Atom as BaseAtom
from enaml.application import deferred_call, Application


class Atom(BaseAtom):

    def safe_set(self, attr, value):
        app = Application.instance()
        if app is None or app.is_main_thread():
            setattr(self, attr, value)
        else:
            deferred_call(setattr, self, attr, value)


def on_thread(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        thread = threading.Thread(target=func,
                                  args=args,
                                  kwargs=kwargs,
                                  daemon=True)
        thread.daemon = True
        thread.start()

    return wrapped



from typing import Iterator, Callable

from enaml.widgets.main_window import MainWindow
from enaml.widgets.widget import Widget
from enaml.widgets.window import Window


def get_window(obj) -> Window:
    for parent in iter_parents(obj):
        if isinstance(parent, Window):
            return parent


def get_main_window(obj) -> MainWindow:
    for parent in iter_parents(obj):
        if isinstance(parent, MainWindow):
            return parent


def iter_parents(obj) -> Iterator[Widget]:
    while (parent := obj.parent) is not None:
        yield parent
        obj = parent


def iter_children(obj, type_=None):
    for child in getattr(obj, 'children', []):
        if type_ is None or isinstance(child, type_):
            yield child
        yield from iter_children(child, type_=type_)


def register_shortcut(obj, key_sequence: str, callback: Callable[[], None]):
    from qtpy.QtWidgets import QShortcut
    from qtpy.QtGui import QKeySequence

    widget = obj.proxy.widget

    shortcut = QShortcut(QKeySequence(key_sequence), widget)

    shortcuts = getattr(widget, '_shortcuts', [])
    shortcuts.append(shortcut)
    widget._shortcuts = shortcuts

    shortcut.activated.connect(callback)
    return shortcut


def register_widget_shortcut(obj, key_sequence, callback):
    from qtpy.QtCore import Qt

    shortcut = register_shortcut(obj, key_sequence, callback)
    shortcut.setContext(Qt.WidgetWithChildrenShortcut)
    return shortcut


def register_window_shortcut(obj, key_sequence, callback):
    from qtpy.QtCore import Qt

    shortcut = register_shortcut(get_window(obj), key_sequence, callback)
    shortcut.setContext(Qt.WindowShortcut)
    return shortcut


def register_application_shortcut(obj, key_sequence, callback):
    from qtpy.QtCore import Qt

    shortcut = register_shortcut(get_main_window(obj), key_sequence, callback)
    shortcut.setContext(Qt.ApplicationShortcut)
    return shortcut



from .table import Table
from .star import StarRating
from .field import Field
from ._shortcuts import ask_yes_no, ask_text

import enaml
with enaml.imports():
    from enamlext.widgets.ipython_console import IPythonConsole


from enaml.qt.qt_factories import QT_FACTORIES

# Injecting the factories to be used by the ProxyResolver
# TODO: should this be here?
def table_factory():
    from enamlext.qt.qt_table import QtTable
    return QtTable


QT_FACTORIES.update(
    Table=table_factory,
)

del QT_FACTORIES
del table_factory


import random
import time
from threading import Thread

import pandas as pd
from atom.api import *

FACTOR = .5


class Model(Atom):
    n = Int()
    i = Int()

    # df1 -> df2
    df_1 = Typed(pd.DataFrame)
    df_2 = Typed(pd.DataFrame)

    @observe('df_1')
    def on_df_1(self, change):
        # print(f'{self.df_1 = }')

        df_2 = self.df_1.copy()
        df_2['value x n'] = df_2['value'] * self.n

        self.df_2 = df_2

    @observe('df_2')
    def on_df_2(self, change):
        print(f'\n\nnew {self.df_2 = }\n\n')

    @observe('n')  # order book tick
    def on_n(self, change):
        # print(f'{self.i}  ->  {self.n}')
        self.i += 1

        if self.df_2 is not None:
            self.df_2['value x n'] = self.df_2['value'] * self.n



def tick(model):
    while True:
        model.n = random.randint(-100_000, 100_000)
        time.sleep(1 * FACTOR)


def update_df(model):
    """Another thread that updates less frequently the dataframe.
    """
    while True:
        fruits = ['apple', 'banana', 'cherry', 'melon', 'pear']
        df = pd.DataFrame()
        picked_fruits = random.sample(fruits, random.randint(1, 5))
        df['fruit'] = picked_fruits
        values = random.sample(range(1, 11), len(df))
        df['value'] = values
        model.df_1 = df
        time.sleep(10 * FACTOR)


def print_df_2(model):
    while True:
        # print(f'\n\n{model.df_2 = }\n{id(model.df_2) = }\n')
        time.sleep(1 * FACTOR)


def main(model=None):
    if model is None:
        model = Model()

    t = Thread(target=tick, args=(model,), daemon=True)
    t.start()

    t2 = Thread(target=update_df, args=(model,), daemon=True)
    t2.start()

    t3 = Thread(target=print_df_2, args=(model,), daemon=True)
    t3.start()

    return t, t2, t3


if __name__ == '__main__':
    t1, t2, t3 = main()
    t1.join()
    t2.join()
    t3.join()



from typing import Any, Dict

from atom.api import Typed, ForwardTyped, List, Bool, observe, Event, Value, Enum
from atom.api import Dict as AtomDict
from atom.atom import set_default
from enaml.core.declarative import d_, d_func
from enaml.widgets.control import Control, ProxyControl

from enamlext.qt.qtable import DoubleClickContext, SelectionContext  # TODO: weak design (leaking Qt details)
from enamlext.qt.table.summary import TableSelectionSummary


ColumnID = str
ColumnKwarg = str
ColumnHint = Dict[ColumnKwarg, Any]
ColumnHints = Dict[ColumnID, ColumnHint]


class ProxyTable(ProxyControl):
    """ The abstract definition of a proxy Table object.
    """
    #: A reference to the Table declaration.
    declaration = ForwardTyped(lambda: Table)

    def set_columns(self, columns):
        raise NotImplementedError

    def set_items(self, items):
        raise NotImplementedError

    def set_selected_items(self, selected_items):
        raise NotImplementedError

    def set_context_menu(self, context_menu):
        raise NotImplementedError

    def set_checkable(self, checkable: bool) -> None:
        raise NotImplementedError

    def set_sortable(self, sortable: bool) -> None:
        raise NotImplementedError

    def refresh_summary(self) -> None:
        raise NotImplementedError

    def set_hints(self, hints: ColumnHints) -> None:
        raise NotImplementedError

    def set_selection_mode(self, selection_mode: str) -> None:
        raise NotImplementedError


class Table(Control):
    """ A tabular grid/table, column-oriented, where individual items are
    displayed in individual rows, and should have the same type.
    """
    #: The columns for the table (horizontal axis/headers)
    columns = d_(List())

    #: The items to be displayed in individual rows of the table
    # items = d_(Instance((Sequence, np.ndarray), factory=list))
    items = d_(Value(factory=list))

    #: Dict of column hints to get passed to Column objects
    # in the case of they are being auto-generated
    hints = d_(AtomDict())

    #: The items that are currently selected on the table # TODO: how to distinguish when mode=cell
    selected_items = d_(List())

    #: Event fired whenever the user double clicks in a cell
    #: The payload will be a DoubleClickContext
    double_clicked = d_(Event(DoubleClickContext), writable=False)

    #: Event fired whenever the selection in the table changes
    selection_changed = d_(Event(SelectionContext), writable=False)

    #: A reference to the ProxyTable object.
    proxy = Typed(ProxyTable)

    # Tables expand freely in height and width by default.
    hug_width = set_default('ignore')
    hug_height = set_default('ignore')

    # Context Menu
    context_menu = d_(List())

    # Flag controlling if the user can check/tick items on the table
    checkable = d_(Bool())

    # Flag controlling if the user can sort items by clicking on a column header on the table
    sortable = d_(Bool(default=True))

    # Flag controlling if the table will have a summary label displayed at the bottom/footer
    show_summary = d_(Bool())

    # Only calculate summary when show_summary is True
    summary = d_(Typed(TableSelectionSummary))

    # Selection mode and behaviour
    selection_mode = d_(Enum('cell', 'cells', 'row', 'rows'))

    # Observers

    @observe("columns",
             "items",
             "selected_items",
             "context_menu",
             "checkable",
             "sortable",
             "show_summary",
             "hints",
             "selection_mode",
             )
    def _update_proxy(self, change: Dict):
        """ An observer which sends state change to the proxy.
        """
        # The superclass handler implementation is sufficient.
        super()._update_proxy(change)

    @d_func
    def refresh(self) -> None:
        if self.initialized:
            self.proxy.widget.refresh()


from functools import cached_property
from typing import Any

from qtpy.QtCore import QModelIndex

def default_convert(item):
    return item

class TableContext:
    """
    Lazy evaluation of properties relative to the data request context.
    The Qt MVC calls the data() in table model, passing the roles...
    Depending on the role, we will call callbacks in the column to retrieve things like
    cell style, font style, tooltip, etc...
    We pass the object of TableContext to the callback, and everything there is
    calculated on demand... which hopefully makes it more efficient
    """

    def __init__(self,
                 model: "QTableModel",
                 index: QModelIndex,
                 role: int,
                 column_index: int,
                 column: "Column",
                 convert = default_convert,
                 ):
        self.__model = model
        self.index = index
        self.role = role
        self.column_index = column_index
        self.column = column
        self.convert = convert

    @cached_property
    def row_index(self) -> int:
        return self.index.row()

    @cached_property
    def item(self) -> Any:
        return self.convert(self._raw_item)

    @cached_property
    def _raw_item(self) -> Any:
        return self.__model.get_item_by_index(self.row_index)

    @cached_property
    def raw_value(self) -> Any:
        return self.column.get_value(self._raw_item)

    @cached_property
    def value(self) -> str:
        """ Returns the displayed value. """
        return self.column.get_displayed_value(self._raw_item)


import enaml


def ask_yes_no(title: str, message: str, parent=None) -> bool:
    with enaml.imports():
        from enaml.stdlib.message_box import question

    button = question(parent, title, message)
    return button is not None and button.text == 'Yes'


def ask_text(title, message, text='', parent=None):
    with enaml.imports():
        from enamlext.widgets._text_input_dialog import TextInputDialog
        dlg = TextInputDialog(parent=parent, title=title, message=message,
                              text=text)
        if dlg.exec_():
            return dlg.text



from atom.api import (
    Typed, ForwardTyped, Int, observe
)
from atom.typing_utils import ChangeDict

from enaml.core.declarative import d_

from enaml.widgets.control import Control, ProxyControl



class ProxyTrafficLight(ProxyControl):
    """ The abstract definition of a proxy TrafficLightWidget object.
    """
    #: A reference to the TrafficLight declaration.
    declaration = ForwardTyped(lambda: TrafficLight)

    def set_color(self, color: str) -> None:
        raise NotImplementedError

    def set_radius(self, radius: int) -> None:
        raise NotImplementedError


class TrafficLight(Control):
    """ Enaml declarative control for giving a TrafficLight widget.
    """
    color = d_(Typed(str))

    radius = d_(Int(default=20))

    #: A reference to the TrafficLight object.
    proxy = Typed(ProxyTrafficLight)

    @observe('color', 'radius')
    def _update_proxy(self, change: ChangeDict) -> None:
        super(TrafficLight, self)._update_proxy(change)


def traffic_light_factory():
    from enamlext.qt.qt_traffic_light import QtTrafficLight
    return QtTrafficLight


from enaml.qt.qt_factories import QT_FACTORIES
QT_FACTORIES['TrafficLight'] = traffic_light_factory


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as fp:
    long_description = fp.read()


setup(
    name='enaml-extensions',
    author='Gabriel Reis',
    author_email='gabrielcnr@gmail.com',
    description='Extra widgets and extensions for building UIs with Enaml',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/gabrielcnr/enaml-extensions',
    license='LGPL',
    keywords='enaml qt widgets extras extensions table qt-table ui gui',
    requires=['enaml'],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    data_files = [("", ["LICENSE"])],
    package_data={
        '': ['*.enaml'],
    },
    include_package_data=True,
    classifiers=[
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Desktop Environment :: Window Managers',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: Software Development :: Widget Sets',
    ],
    use_scm_version={"write_to": "src/enamlext/_version.py"},
    setup_requires=["setuptools-scm", "setuptools>=30.3.0"],    
    python_requires=">=3.10",
    install_requires=[
        "enaml",
        "QScintilla",
        "pip",
        "qtpy",
        "numpy",
    ],
    extras_require={
        "testing": [
            "pytest",
            "pytest-mock",
            "pytest-qt",
            "pandas",
        ]
    },    
    tests_requires=['pytest', 'pytest-mock', 'pytest-qt', 'pandas'],
)



import threading
import time

from atom.api import Int
from enaml.qt.qt_application import QtApplication

from enamlext.atom import Atom, on_thread


def test_safe_set_without_application():
    class Model(Atom):
        value = Int()

    model = Model()
    model.safe_set("value", 12)

    assert 12 == model.value


def test_safe_set_with_application_from_main_ui_thread():
    class Model(Atom):
        value = Int()

    model = Model()

    app = QtApplication()

    model.safe_set("value", 15)
    assert 15 == model.value

    app.destroy()


def test_safe_set_with_application_from_another_thread(mocker):
    mock_deferred_call = mocker.patch("enamlext.atom.deferred_call", autospec=True)

    class Model(Atom):
        value = Int()

    model = Model()

    app = QtApplication()

    def worker(given_model):
        given_model.safe_set("value", 23)

    thread = threading.Thread(target=worker, args=(model,))
    thread.daemon = True
    thread.start()

    time.sleep(.2)

    mock_deferred_call.assert_called_once_with(
        setattr, model, "value", 23
    )

    app.destroy()


def test_on_thread(mocker):
    mock_deferred_call = mocker.patch("enamlext.atom.deferred_call", autospec=True)

    class Model(Atom):
        value = Int()

    model = Model()

    app = QtApplication()

    @on_thread
    def worker(given_model):
        given_model.safe_set("value", 19)

    worker(model)

    time.sleep(.2)

    mock_deferred_call.assert_called_once_with(
        setattr, model, "value", 19
    )

    app.destroy()



#!/usr/bin/env python


#############################################################################
##
## Copyright (C) 2010 Hans-Peter Jansen <hpj@urpla.net>.
## Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
## All rights reserved.
##
## This file is part of the examples of PyQt.
##
## $QT_BEGIN_LICENSE:BSD$
## You may use this file under the terms of the BSD license as follows:
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in
##     the documentation and/or other materials provided with the
##     distribution.
##   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
##     the names of its contributors may be used to endorse or promote
##     products derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
## $QT_END_LICENSE$
##
#############################################################################


# These are only needed for Python v2 but are harmless for Python v3.
import sip

sip.setapi('QString', 2)
sip.setapi('QVariant', 2)

import math

from PyQt4 import QtCore, QtGui
from PyQt4.QtGui import *

from atom.api import Typed
from enamlext.widgets.star import ProxyStarRating
from enaml.qt.qt_control import QtControl


class StarRating(object):
    # enum EditMode
    Editable, ReadOnly = range(2)

    PaintingScaleFactor = 20

    def __init__(self, starCount=1, maxStarCount=5):
        self._starCount = starCount
        self._maxStarCount = maxStarCount

        self.starPolygon = QtGui.QPolygonF([QtCore.QPointF(1.0, 0.5)])
        for i in range(5):
            self.starPolygon << QtCore.QPointF(
                0.5 + 0.5 * math.cos(0.8 * i * math.pi),
                0.5 + 0.5 * math.sin(0.8 * i * math.pi))

        self.diamondPolygon = QtGui.QPolygonF()
        self.diamondPolygon << QtCore.QPointF(0.4, 0.5) \
        << QtCore.QPointF(0.5, 0.4) \
        << QtCore.QPointF(0.6, 0.5) \
        << QtCore.QPointF(0.5, 0.6) \
        << QtCore.QPointF(0.4, 0.5)

    def starCount(self):
        return self._starCount

    def maxStarCount(self):
        return self._maxStarCount

    def setStarCount(self, starCount):
        print 'setStarCount', starCount
        self._starCount = starCount

    def setMaxStarCount(self, maxStarCount):
        self._maxStarCount = maxStarCount

    def sizeHint(self):
        return self.PaintingScaleFactor * QtCore.QSize(self._maxStarCount, 1)

    def paint(self, painter, rect, palette, editMode):
        painter.save()

        painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.setPen(QColor('#DAA520'))

        painter.setBrush(QColor('#DAA520'))
        # if editMode == StarRating.Editable:
        #     painter.setBrush(palette.background())
        # else:
        #     painter.setBrush(palette.foreground())

        yOffset = (rect.height() - self.PaintingScaleFactor) / 2
        painter.translate(rect.x(), rect.y() + yOffset)
        painter.scale(self.PaintingScaleFactor, self.PaintingScaleFactor)

        for i in range(self._maxStarCount):
            if i < self._starCount:
                painter.drawPolygon(self.starPolygon, QtCore.Qt.WindingFill)
            elif editMode == StarRating.Editable:
                painter.drawPolygon(self.diamondPolygon, QtCore.Qt.WindingFill)

            painter.translate(1.0, 0.0)

        painter.restore()


class StarEditor(QtGui.QWidget):
    editingFinished = QtCore.pyqtSignal()
    ratingChanged = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super(StarEditor, self).__init__(parent)

        self._starRating = StarRating()

        self.setMouseTracking(True)
        self.setAutoFillBackground(True)

    def setStarRating(self, starRating):
        self._starRating = starRating

    def starRating(self):
        return self._starRating

    def sizeHint(self):
        return self._starRating.sizeHint()

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        self._starRating.paint(painter, self.rect(), self.palette(),
                               StarRating.Editable)

    def mouseReleaseEvent(self, event):
        star = self.starAtPosition(event.x())

        if star != self._starRating.starCount() and star != -1:
            self._starRating.setStarCount(star)
            self.update()
        self.ratingChanged.emit()
        self.editingFinished.emit()

    def starAtPosition(self, x):
        # Enable a star, if pointer crosses the center horizontally.
        starwidth = self._starRating.sizeHint().width() // self._starRating.maxStarCount()
        star = (x + starwidth / 2) // starwidth
        if 0 <= star <= self._starRating.maxStarCount():
            return star

        return -1


class StarDelegate(QtGui.QStyledItemDelegate):
    def paint(self, painter, option, index):
        starRating = index.data()
        if isinstance(starRating, StarRating):
            if option.state & QtGui.QStyle.State_Selected:
                painter.fillRect(option.rect, option.palette.highlight())

            starRating.paint(painter, option.rect, option.palette,
                             StarRating.ReadOnly)
        else:
            super(StarDelegate, self).paint(painter, option, index)

    def sizeHint(self, option, index):
        starRating = index.data()
        if isinstance(starRating, StarRating):
            return starRating.sizeHint()
        else:
            return super(StarDelegate, self).sizeHint(option, index)

    def createEditor(self, parent, option, index):
        starRating = index.data()
        if isinstance(starRating, StarRating):
            editor = StarEditor(parent)
            editor.editingFinished.connect(self.commitAndCloseEditor)
            return editor
        else:
            return super(StarDelegate, self).createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        starRating = index.data()
        if isinstance(starRating, StarRating):
            editor.setStarRating(starRating)
        else:
            super(StarDelegate, self).setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        starRating = index.data()
        if isinstance(starRating, StarRating):
            model.setData(index, editor.starRating())
        else:
            super(StarDelegate, self).setModelData(editor, model, index)

    def commitAndCloseEditor(self):
        editor = self.sender()
        self.commitData.emit(editor)
        self.closeEditor.emit(editor)


class QtStarRating(QtControl, ProxyStarRating):
    """ A Qt implementation of an Enaml ProxyStarRating.

    """
    #: A reference to the widget created by the proxy.
    widget = Typed(StarEditor)

    # --------------------------------------------------------------------------
    # Initialization API
    # --------------------------------------------------------------------------
    def create_widget(self):
        """ Create the underlying StarEditor widget.

        """
        self.widget = StarEditor(self.parent_widget())

    def init_widget(self):
        """ Initialize the underlying widget.

        """
        super(QtStarRating, self).init_widget()
        d = self.declaration
        self.set_stars(d.stars)
        self.set_rating(d.rating)
        self.widget.ratingChanged.connect(self._on_rating_changed)

    def _on_rating_changed(self):
        self.declaration.rating = self.widget.starRating().starCount()

    # --------------------------------------------------------------------------
    # ProxyStarRating API
    # --------------------------------------------------------------------------
    def set_stars(self, stars):
        self.widget.starRating().setMaxStarCount(stars)
        self.widget.repaint()

    def set_rating(self, rating):
        if rating > self.declaration.stars:
            self.declaration.rating = self.declaration.stars
            return
        elif rating < 0:
            self.declaration.rating = 0
            return
        self.widget.starRating().setStarCount(rating)
        self.widget.repaint()



from atom.api import Atom, Value, Typed
from enaml.widgets.api import Form, Label, DatetimeSelector
from enaml.core.declarative import d_
from enamlext.widgets import Field
import datetime
import enaml
with enaml.imports():
    from enaml.stdlib.fields import IntField


def connect(model, model_attr, view, view_attr):
    assert isinstance(model, Atom)
    assert isinstance(view, Atom)

    model.observe(model_attr,
                  lambda change: setattr(view, view_attr, change['value']))
    view.observe(view_attr,
                 lambda change: setattr(model, model_attr, change['value']))

    # Initially the view should be sync'd with the model
    try:
        setattr(view, view_attr, getattr(model, model_attr))
    except Exception as ex:
        #import pdb ;pdb.set_trace()
        raise


def create_editor(obj, form=None):
    assert isinstance(obj, Atom)

    klass = obj.__class__

    form = form or Form()

    # get the members
    for member_name, member_descriptor in klass.members().iteritems():
        # by default we are skipping members starting with _
        if member_name.startswith('_'):
            continue

        m = member_descriptor.metadata or {}
        if m.get('exclude', False):
            continue

        label_text = m.get('label', member_name)
        label = Label(text=label_text)

        widget = Field()
        model_attr = 'text'
        if isinstance(member_descriptor, Typed):
            type_ = member_descriptor.validate_mode[1]
            if issubclass(type_, datetime.date):
                widget = DatetimeSelector(calendar_popup=True)
                model_attr = 'datetime'
            elif issubclass(type_, int):
                widget = IntField()
                model_attr = 'value'

        form.insert_children(None, [label, widget])

        connect(obj, member_name, widget, model_attr)

    return form


class Editor(Form):
    obj = d_(Value())

    def __init__(self, *args, **kwargs):
        super(Editor, self).__init__(*args, **kwargs)
        create_editor(self.obj, self)
